from __future__ import annotations
import inspect
from typing import Callable, Iterable, Union, TYPE_CHECKING, TypeVar, overload
import warnings
from magicgui.widgets import FunctionGui

from .utils import show_messagebox
from .types import Color
from .signature import upgrade_signature

if TYPE_CHECKING:
    from ._gui import BaseGui

nStrings = Union[str, Iterable[str]]

R = TypeVar("R")
T = TypeVar("T")
F = TypeVar("F", bound=Callable)


def set_options(
    layout: str = "vertical",
    labels: bool = True,
    call_button: bool | str | None = None,
    auto_call: bool = False,
    **options,
) -> Callable[[F], F]:
    """
    Set MagicSignature to functions.

    By decorating a method with this function, ``magicgui`` will create a widget with these
    options. These codes are similar in appearance.

    .. code-block:: python

        # A magicgui way
        @magicgui(a={...})
        def func(a):
            ...

        # A magicclass way
        @magicclass
        class A:
            @set_options(a={...})
            def func(self, a):
                ...

    Parameters
    ----------
    layout : str, default is "vertical"
        The type of layout to use in FunctionGui. Must be one of {'horizontal', 'vertical'}.
    labels : bool, default is True
        Whether labels are shown in the FunctionGui.
    call_button : bool or str, optional
        If ``True``, create an additional button that calls the original
        function when clicked.  If a ``str``, set the button text. If None (the
        default), it defaults to True when ``auto_call`` is False, and False
        otherwise.
    auto_call : bool, optional
        If ``True``, changing any parameter in either the GUI or the widget attributes
        will call the original function with the current settings. by default False
    options : dict
        Parameter options.
    """

    def wrapper(func: F) -> F:
        sig = inspect.signature(func)
        rem = options.keys() - sig.parameters.keys()
        if rem:
            warnings.warn(
                f"Unknown arguments found in set_options of {func.__name__}: {rem}",
                UserWarning,
            )

        upgrade_signature(
            func,
            gui_options=options,
            additional_options={
                "call_button": call_button,
                "layout": layout,
                "labels": labels,
                "auto_call": auto_call,
            },
        )

        return func

    return wrapper


def set_design(
    width: int | None = None,
    height: int | None = None,
    min_width: int | None = None,
    min_height: int | None = None,
    max_width: int | None = None,
    max_height: int | None = None,
    text: str | None = None,
    icon: str | None = None,
    icon_path: str | None = None,
    font_size: int | None = None,
    font_family: int | None = None,
    font_color: Color | None = None,
    background_color: Color | None = None,
    visible: bool | None = None,
) -> Callable[[type[T]], type[T]] | Callable[[F], F]:
    """
    Change button/action design by calling setter when the widget is created.

    Parameters
    ----------
    width : int, optional
        Button width. Call ``button.width = width``.
    height : int, optional
        Button height. Call ``button.height = height``.
    min_width : int, optional
        Button minimum width. Call ``button.min_width = min_width``.
    min_height : int, optional
        Button minimum height. Call ``button.min_height = min_height``.
    max_width : int, optional
        Button maximum width. Call ``button.max_width = max_width``.
    max_height : int, optional
        Button maximum height. Call ``button.max_height = max_height``.
    text : str, optional
        Button text. Call ``button.text = text``.
    icon : str, optional
        Path to icon file. ``min_width`` and ``min_height`` will be automatically set to the icon size
        if not given.
    font_size : int, optional
        Font size of the text.
    visible : bool default is True
        Button visibility.
    """
    caller_options = locals()
    caller_options = {k: v for k, v in caller_options.items() if v is not None}

    if "icon_path" in caller_options.keys():
        warnings.warn(
            "`icon_path` is deprecated. Use `icon` instead.", DeprecationWarning
        )
        caller_options["icon"] = caller_options.pop("icon_path")

    def wrapper(obj):
        if isinstance(obj, type):
            _post_init = getattr(obj, "__post_init__", lambda self: None)

            def __post_init__(self):
                _post_init(self)
                for k, v in caller_options.items():
                    setattr(self, k, v)

            obj.__post_init__ = __post_init__
        else:
            upgrade_signature(obj, caller_options=caller_options)
        return obj

    return wrapper


def do_not_record(method: F) -> F:
    """Wrapped method will not be recorded in macro."""
    upgrade_signature(method, additional_options={"record": False})
    return method


def bind_key(*key) -> Callable[[F], F]:
    """
    Define a keybinding to a button or an action.
    This function accepts several styles of shortcut expression.

    >>> @bind_key("Ctrl-A")         # napari style
    >>> @bind_key("Ctrl", "A")      # separately
    >>> @bind_key(Key.Ctrl + Key.A) # use Key class
    >>> @bind_key(Key.Ctrl, Key.A)  # use Key class separately

    """
    if isinstance(key[0], tuple):
        key = key[0]

    def wrapper(method: F) -> F:
        upgrade_signature(method, additional_options={"keybinding": key})
        return method

    return wrapper


class Canceled(RuntimeError):
    """Raised when a function is canceled"""


@overload
def confirm(
    *,
    text: str | None,
    condition: Callable[..., bool] | str | None,
    callback: Callable[[str, BaseGui], None] | None = None,
) -> Callable[[F], F]:
    ...


@overload
def confirm(
    f: F,
    *,
    text: str | None,
    condition: Callable[..., bool] | str | None,
    callback: Callable[[str, BaseGui], None] | None = None,
) -> F:
    ...


def confirm(
    f: F | None = None,
    *,
    text: str | None = None,
    condition: Callable[[BaseGui], bool] | str = None,
    callback: Callable[[str, BaseGui], None] | None = None,
):
    """
    Confirm if it is OK to run function in GUI.

    Useful when the function will irreversibly delete or update something in GUI.
    Confirmation will be executed only when function is called in GUI.

    Parameters
    ----------
    text : str, optional
        Confirmation message, such as "Are you sure to run this function?". Format
        string can also be used here, in which case arguments will be passed. For
        instance, to execute confirmation on function ``f(a, b)``, you can use
        format string ``"Running with a = {a} and b = {b}"`` then confirmation
        message will be "Running with a = 1, b = 2" if ``f(1, 2)`` is called.
        By default, message will be "Do you want to run {name}?" where "name" is
        the function name.
    condition : callable or str, optional
        Condition of when confirmation will show up. If callable, it must accept
        ``condition(self)`` and return boolean object. If string, it must be
        evaluable as literal with input arguments as local namespace. For instance,
        function ``f(a, b)`` decorated by ``confirm(condition="a < b + 1")`` will
        evaluate ``a < b + 1`` to check if confirmation is needed. Always true by
        default.
    callback : callable, optional
        Callback function when confirmation is needed. Must take a ``str`` and a
        ``BaseGui`` object as inputs. By default, message box will be shown. Useful
        for testing.
    """
    if condition is None:
        condition = lambda x: True
    if callback is None:
        callback = _default_confirmation

    def _decorator(method: F) -> F:
        _name = method.__name__

        # set text
        if text is None:
            _text = f"Do you want to run {_name}?"
        elif isinstance(text, str):
            _text = text
        else:
            raise TypeError(
                f"The first argument of 'confirm' must be a str but got {type(text)}."
            )
        upgrade_signature(
            method,
            additional_options={
                "confirm": {
                    "text": _text,
                    "condition": condition,
                    "callback": callback,
                }
            },
        )
        return method

    if f is not None:
        return _decorator(f)
    return _decorator


def _default_confirmation(text: str, gui: BaseGui):
    ok = show_messagebox(
        mode="question",
        title="Confirmation",
        text=text,
        parent=gui.native,
    )
    if not ok:
        raise Canceled("Canceled")


def nogui(method: F) -> F:
    """Wrapped method will not be converted into a widget."""
    upgrade_signature(method, additional_options={"gui": False})
    return method


def mark_preview(function: Callable, text: str = "Preview") -> Callable[[F], F]:
    """
    Define a preview of a function.

    This decorator is useful for advanced magicgui creation. A "Preview" button
    appears in the bottom of the widget built from the input function and the
    decorated function will be called with the same arguments.
    Following example shows how to define a previewer that prints the content of
    the selected file.

    .. code-block:: python

        def func(self, path: Path):
            ...

        @mark_preview(func)
        def _func_prev(self, path: Path):
            with open(path, mode="r") as f:
                print(f.read())

    Parameters
    ----------
    function : callable
        To which function previewer will be defined.
    text : str, optional
        Text of preview button.
    """

    def _wrapper(preview: F) -> F:
        sig_preview = inspect.signature(preview)
        sig_func = inspect.signature(function)
        params_preview = sig_preview.parameters
        params_func = sig_func.parameters

        less = len(params_func) - len(params_preview)
        if less == 0:
            if params_preview.keys() != params_func.keys():
                raise TypeError(
                    f"Arguments mismatch between {sig_preview!r} and {sig_func!r}."
                )
            # If argument names are identical, input arguments don't have to be filtered.
            _filter = lambda a: a

        elif less > 0:
            idx: list[int] = []
            for i, param in enumerate(params_func.keys()):
                if param in params_preview:
                    idx.append(i)
            # If argument names are not identical, input arguments have to be filtered so
            # that arguments match the inputs.
            _filter = lambda _args: (a for i, a in enumerate(_args) if i in idx)

        else:
            raise TypeError(
                f"Number of arguments of preview function {preview!r} must be equal"
                f"or smaller than that of running function {function!r}."
            )

        def _preview(*args):
            # find proper parent instance in the case of classes being nested
            from ._gui import BaseGui

            if len(args) > 0 and isinstance(args[0], BaseGui):
                ins = args[0]
                prev_ns = preview.__qualname__.split(".")[-2]
                while ins.__class__.__name__ != prev_ns:
                    ins = ins.__magicclass_parent__
                args = (ins,) + args[1:]
            # filter input arguments
            return preview(*_filter(args))

        if not isinstance(function, FunctionGui):
            upgrade_signature(
                function, additional_options={"preview": (text, _preview)}
            )
        else:
            from ._gui._function_gui import append_preview

            append_preview(function, _preview, text=text)
        return preview

    return _wrapper



from .utils import to_napari
from .widgets import NapariCanvas
from .viewer import ViewerWidget
from ._magicgui import _register_mgui_types

_register_mgui_types()

del _register_mgui_types

__all__ = ["to_napari", "NapariCanvas", "ViewerWidget"]



from __future__ import annotations
from typing import Any, Iterable, MutableSequence
from qtpy.QtWidgets import QListWidget, QListWidgetItem, QAbstractItemView

from .object import BaseWidget, ContextMenuMixin, PyObjectBound


class ListWidget(BaseWidget, MutableSequence):
    def __init__(self, value: Iterable[Any] = None, dragdrop: bool = True, **kwargs):
        """
        A widget composed of a list of items. This widget behaves very similar to Python
        list so that you can append, insert, pop or do any indexing.

        Parameters
        ----------
        value : Iterable, optional
            Initial value of the list widget. Any iterable object that can be passed to
            Python list.
        dragdrop : bool, default is True
            Allow drag and drop of list contents.
        """
        super().__init__(**kwargs)
        self._listwidget = PyListWidget(self.native)
        self._listwidget.setParentWidget(self)
        self.set_widget(self._listwidget)

        @self._listwidget.itemDoubleClicked.connect
        def _(item: PyListWidgetItem):
            type_ = type(item.obj)
            callbacks = self._callbacks.get(type_, [])
            self.running = True
            try:
                for callback in callbacks:
                    try:
                        callback(item.obj, self._listwidget.row(item))
                    except TypeError:
                        callback(item.obj)
            finally:
                self.running = False

        if dragdrop:
            self._listwidget.setAcceptDrops(True)
            self._listwidget.setDragEnabled(True)
            self._listwidget.setDragDropMode(
                QAbstractItemView.DragDropMode.InternalMove
            )

        if value is not None:
            for v in value:
                self.append(v)

    def __len__(self) -> int:
        """
        Length of widget contents.
        """
        return self._listwidget.count()

    @property
    def value(self) -> list[Any]:
        """
        Get all the contents as a Python list.

        Returns
        -------
        list
            Contents of the list widget.
        """
        return [self._listwidget.item(i).obj for i in range(len(self))]

    def insert(self, i: int, obj: Any):
        """
        Insert object of any type to the list.
        """
        name = self._delegates.get(type(obj), str)(obj)
        tooltip = self._tooltip.get(type(obj), str)(obj)
        item = PyListWidgetItem(self._listwidget, obj=obj, name=name)
        item.setToolTip(tooltip)
        self._listwidget.insertItem(i, item)

    def __getitem__(self, i: int) -> Any:
        """
        Get i-th Python object (not widget item object!).
        """
        return self._listwidget.item(i).obj

    def __setitem__(self, i: int, obj: Any):
        """
        Set i-th Python object.
        """
        self._listwidget.takeItem(i)
        self.insert(i, obj)

    def __delitem__(self, i: int):
        """
        Delete i-th Python object from the list widget.
        """
        self._listwidget.takeItem(i)

    def clear(self):
        """
        Clear all the items.
        """
        self._listwidget.clear()

    def index(self, obj: Any, start: int = 0, stop: int = None) -> int:
        """
        Find object or list widget item from the list widget.

        Parameters
        ----------
        obj : Any
            Object to find. If a PyListWidgetItem is given, the index of the item
            (not the tagged object) is searched for.
        start : int, optional
            Starting index, by default 0
        stop : int, optional
            Index to stop searching.

        Returns
        -------
        int
            Index of object.

        Raises
        ------
        ValueError
            If object was not found.
        """
        if isinstance(obj, PyListWidgetItem):
            f = self._listwidget.item
        else:
            f = lambda i: self._listwidget.item(i).obj
        if start is not None and start < 0:
            start = max(len(self) + start, 0)
        if stop is not None and stop < 0:
            stop += len(self)

        i = start
        while stop is None or i < stop:
            try:
                v = f(i)
                if v is obj or v == obj:
                    return i
            except IndexError:
                break
            i += 1
        raise ValueError(f"Object {obj!r} not found in the list")


class PyListWidget(ContextMenuMixin, QListWidget):
    def item(self, row: int) -> PyListWidgetItem:
        return super().item(row)

    def itemAt(self, *p) -> PyListWidgetItem:
        return super().itemAt(*p)

    def __init__(self, parent: None) -> None:
        super().__init__(parent=parent)
        self.setContextMenu()


class PyListWidgetItem(PyObjectBound, QListWidgetItem):
    def __init__(self, parent: QListWidget = None, obj=None, name=None):
        super().__init__(parent)
        self.setObject(obj, name)



from magicclass import magicclass, magictoolbar, magicmenu, MagicTemplate
from unittest.mock import MagicMock
from pathlib import Path
from types import MethodType

from magicclass.fields import field

def test_basic():
    @magicclass
    class A(MagicTemplate):
        def f1(self): ...
        def f2(self, path: Path): ...
        # "type" will raise error if magicgui try to interpret the annotation
        def _private(self, arg: type): ...
        def f3(self, i: int): ...

    ui = A()
    assert len(ui) == 3
    assert ui[0].name == "f1"
    assert ui[1].name == "f2"
    assert ui[2].name == "f3"
    assert isinstance(ui["_private"], MethodType)

    # test macro
    mock = MagicMock()
    ui.changed.connect(mock)
    ui[0].changed()
    assert str(ui.macro[-1]) == "ui.f1()"
    mock.assert_called()


def test_menu():
    @magicclass
    class A:
        @magicmenu
        class Menu:
            def m1(self): ...
            def m2(self, path: Path): ...
            @magicmenu
            class Inner:
                def inner1(self): ...
                def _private(self): ...
                def inner2(self): ...
            def m3(self, i: int): ...
        def f1(self): ...
        def f2(self, path: Path): ...
        def f3(self, i: int): ...

    ui = A()
    assert len(ui) == 3
    assert ui._menubar is not None
    assert len(ui.Menu) == 4
    assert len(ui.Menu.Inner) == 2

    # test macro
    ui["f1"].changed()
    assert str(ui.macro[-1]) == "ui.f1()"
    ui.Menu.Inner["inner1"].changed()
    assert str(ui.macro[-1]) == "ui.Menu.Inner.inner1()"

def test_toolbar():
    @magicclass
    class A:
        @magictoolbar
        class Toolbar:
            def m1(self): ...
            def m2(self, path: Path): ...
            @magicmenu
            class Menu:
                def inner1(self): ...
                def _private(self): ...
                def inner2(self): ...
            def m3(self, i: int): ...
        def f1(self): ...
        def f2(self, path: Path): ...
        def f3(self, i: int): ...

    ui = A()
    assert len(ui) == 3
    assert ui._menubar is None
    assert len(ui.Toolbar) == 4
    assert len(ui.Toolbar.Menu) == 2

    # test macro
    ui["f1"].changed()
    assert str(ui.macro[-1]) == "ui.f1()"
    ui.Toolbar.Menu["inner1"].changed()
    assert str(ui.macro[-1]) == "ui.Toolbar.Menu.inner1()"


def test_separator():
    from magicclass.widgets import Separator
    @magicclass
    class A:
        @magicmenu
        class Menu:
            def m1(self): ...
            sep0 = field(Separator)
            i = field(int)
            sep1 = Separator()
            def m2(self, path: Path): ...

    ui = A()
    assert type(ui.Menu[1].widget) is Separator
    assert type(ui.Menu[3].widget) is Separator

    @magicclass
    class A:
        @magictoolbar
        class Menu:
            def m1(self): ...
            sep0 = field(Separator)
            i = field(int)
            sep1 = Separator()
            def m2(self, path: Path): ...

    ui = A()
    assert type(ui.Menu[1].widget) is Separator
    assert type(ui.Menu[3].widget) is Separator


def test_tooltip():
    @magicclass(widget_type="mainwindow")
    class A:
        @magicmenu
        class Menu:
            def func(self, a: int):
                """
                Test tooltip.

                Parameters
                ----------
                a : int
                    Tooltip of a.
                """

        @magictoolbar
        class Tool:
            def func(self, a: int):
                """
                Test tooltip.

                Parameters
                ----------
                a : int
                    Tooltip of a.
                """

        def func(self, a: int):
            """
            Test tooltip.

            Parameters
            ----------
            a : int
                Tooltip of a.
            """

    _TOOLTIP = "Test tooltip."
    _PARAM_TOOLTIP = "Tooltip of a."

    ui = A()
    ui.show(False)

    assert ui["func"].tooltip == _TOOLTIP
    assert ui.Menu["func"].tooltip == _TOOLTIP
    assert ui.Tool["func"].tooltip == _TOOLTIP

    ui["func"].changed()
    ui["func"].mgui._call_button.changed()
    ui.Menu["func"].changed()
    ui.Menu["func"].mgui._call_button.changed()
    ui.Tool["func"].changed()
    ui.Tool["func"].mgui._call_button.changed()

    assert ui["func"].mgui.a.tooltip == _PARAM_TOOLTIP
    assert ui.Menu["func"].mgui.a.tooltip == _PARAM_TOOLTIP
    assert ui.Tool["func"].mgui.a.tooltip == _PARAM_TOOLTIP


def test_post_append():
    @magicclass(widget_type="mainwindow")
    class A:
        @magicmenu
        class Menu:
            def func(self, a: int): ...

        @magictoolbar
        class Tool:
            def func(self, a: int): ...

    ui = A()

    def new_func_1(): ...
    def new_func_2(): ...
    def new_func_3(): ...

    ui.append(new_func_1)
    ui.Menu.append(new_func_2)
    ui.Tool.append(new_func_3)

    ui["new_func_1"].changed()
    assert str(ui.macro[-1]) == "ui.new_func_1()"
    ui.Menu["new_func_2"].changed()
    assert str(ui.macro[-1]) == "ui.Menu.new_func_2()"
    ui.Tool["new_func_3"].changed()
    assert str(ui.macro[-1]) == "ui.Tool.new_func_3()"


def test_labels_arg():
    """Test 'labels=False' works in menubar and toolbar."""
    @magicclass
    class A:
        @magicmenu
        class Menu:
            a = field(int)

        @magictoolbar
        class Tool:
            a = field(int)

    ui = A()

    assert ui.Menu.a._labeled_widget() is not None
    assert ui.Tool.a._labeled_widget() is not None

    @magicclass
    class A:
        @magicmenu(labels=False)
        class Menu:
            a = field(int)

        @magictoolbar(labels=False)
        class Tool:
            a = field(int)

    ui = A()

    assert ui.Menu.a._labeled_widget() is None
    assert ui.Tool.a._labeled_widget() is None


@magicmenu(name="Menu 1")
class Menu1:
    def m(self): ...

@magicclass(name="My Widget")
class MyWidget:
    def f(self): ...

@magicclass
class Main:
    @magicmenu
    class Menu:
        menu1 = Menu1
    MyWidget = MyWidget

def test_names():
    ui = Main()
    assert ui.MyWidget.name == "My Widget"
    assert ui.Menu.menu1.name == "Menu 1"
    ui.MyWidget["f"].changed()
    assert str(ui.macro[-1]) == "ui.MyWidget.f()"
    ui.Menu.menu1["m"].changed()
    assert str(ui.macro[-1]) == "ui.Menu.menu1.m()"

def test_default_init():
    @magicclass
    class A:
        pass

    ui_a = A(name="a")
    ui_b = A(name="b")
    assert ui_a.name == "a"
    assert ui_b.name == "b"

    @magicclass
    class B:
        def __init__(self, param=None):
            self.param = param

    ui = B(param=10)
    assert ui.param == 10



from __future__ import annotations

from typing import TYPE_CHECKING, Annotated, Iterable, Tuple, List

# for now...
if TYPE_CHECKING:
    import pandas as pd

    Features = pd.DataFrame
    FeatureColumn = pd.Series
else:
    from typing import Protocol, Iterable

    class Features(Protocol):
        @property
        def columns(self) -> Iterable[str]:
            ...

    class FeatureColumn:
        pass


__all__ = ["Features", "FeatureColumn", "FeatureInfo"]


class _FeatureInfoAlias(type):
    def __getitem__(cls, names: str | tuple[str, ...]):
        if isinstance(names, str):
            names = (names,)
        else:
            for name in names:
                if not isinstance(name, str):
                    raise ValueError(
                        f"Cannot subscribe type {type(name)} to FeatureInfo."
                    )

        return Annotated[FeatureInfoInstance, {"column_choice_names": names}]


class FeatureInfo(metaclass=_FeatureInfoAlias):
    """
    A type representing a tuple of a DataFrame and its column names.

    .. code-block:: python

        def func(x: FeatureInfo["x", "y"]): ...

    Type annotation ``FeatureInfo["x", "y"]`` is essentially equivalent to
    ``tuple[pd.DataFrame, tuple[str, str]]``.
    """

    def __new__(cls, *args, **kwargs):
        raise TypeError(f"Type {cls.__name__} cannot be instantiated.")


class FeatureInfoInstance(Tuple["pd.DataFrame", List[str]]):
    def __new__(cls, *args, **kwargs):
        raise TypeError(f"Type {cls.__name__} cannot be instantiated.")



from __future__ import annotations
from typing import Any, Iterable, MutableMapping
from qtpy.QtWidgets import QTableWidget, QTableWidgetItem

from .object import BaseWidget, ContextMenuMixin, PyObjectBound


class DictWidget(BaseWidget, MutableMapping):
    def __init__(self, value=None, **kwargs):
        super().__init__(**kwargs)

        self._tablewidget = PyTableWidget(self.native)
        self._tablewidget.setParentWidget(self)
        self._tablewidget.setEditTriggers(QTableWidget.NoEditTriggers)
        self._tablewidget.verticalHeader().setDefaultSectionSize(30)
        self._dict: dict[str, int] = {}  # mapping from key to row

        self.set_widget(self._tablewidget)

        @self._tablewidget.itemDoubleClicked.connect
        def _(item: PyTableWidgetItem):
            type_ = type(item.obj)
            callbacks = self._callbacks.get(type_, [])
            self.running = True
            try:
                for callback in callbacks:
                    try:
                        callback(item.obj, self._tablewidget.row(item))
                    except TypeError:
                        callback(item.obj)
            finally:
                self.running = False

        if value is not None:
            self.update(dict(value))

    def __len__(self) -> int:
        return self._tablewidget.rowCount()

    @property
    def value(self) -> dict[str, Any]:
        return {k: self._tablewidget.item(row, 0) for k, row in self._dict}

    def __getitem__(self, k: str) -> Any:
        row = self._dict[k]
        return self._tablewidget.item(row, 0).obj

    def __setitem__(self, k: str, obj: Any) -> None:
        if not isinstance(k, str):
            raise ValueError("Can only use str type as keys.")

        if k in self._dict.keys():
            row = self._dict[k]
        else:
            row = len(self)
            self._dict[k] = row
            self._tablewidget.insertRow(row)
            if row == 0:
                self._tablewidget.insertColumn(0)
                self._tablewidget.setHorizontalHeaderItem(0, QTableWidgetItem("value"))
            key_item = QTableWidgetItem(k)
            self._tablewidget.setVerticalHeaderItem(row, key_item)

        name = self._delegates.get(type(obj), str)(obj)
        value_item = PyTableWidgetItem(obj, name)
        tooltip = self._tooltip.get(type(obj), str)(obj)
        value_item.setToolTip(tooltip)
        self._tablewidget.setItem(row, 0, value_item)

    def __delitem__(self, k: str) -> None:
        row = self._dict.pop(k)
        self._tablewidget.removeRow(row)

    def __iter__(self) -> Iterable[str]:
        return iter(self._dict)

    def keys(self):
        """
        Return the view of dictionary keys.
        """
        return self._dict.keys()

    def values(self) -> DictValueView:
        """
        Return the view of dictionary values as Python objects.
        """
        return DictValueView(self._tablewidget)

    def items(self) -> DictItemView:
        """
        Return the view of dictionary keys and values as strings and Python objects.
        """
        return DictItemView(self._tablewidget)

    def update(self, d: dict[str, Any]):
        """
        Update the dictionary contents.
        """
        for k, v in d.items():
            self[k] = v

    def clear(self) -> None:
        """
        Clear dictionary contents.
        """
        self._tablewidget.clear()
        self._dict.clear()

    def pop(self, k: str):
        """
        Pop a dictionary content.
        """
        row = self._dict.pop(k)
        out = self._tablewidget.item(row, 0).obj
        self._tablewidget.removeRow(row)
        return out

    def get(self, k: str, default=None):
        self._dict.get(k, default)


class PyTableWidget(ContextMenuMixin, QTableWidget):
    def item(self, row: int, column: int) -> PyTableWidgetItem:
        return super().item(row, column)

    def itemAt(self, *p) -> PyTableWidgetItem:
        return super().itemAt(*p)

    def __init__(self, parent: None) -> None:
        super().__init__(parent=parent)
        self.setContextMenu()


class PyTableWidgetItem(PyObjectBound, QTableWidgetItem):
    def __init__(self, obj=None, name=None):
        super().__init__()
        self.setObject(obj, name)


class DictValueView:
    def __init__(self, widget: PyTableWidget):
        self.widget = widget

    def __iter__(self):
        for row in range(self.widget.rowCount()):
            yield self.widget.item(row, 0).obj


class DictItemView:
    def __init__(self, widget: PyTableWidget):
        self.widget = widget

    def __iter__(self):
        for row in range(self.widget.rowCount()):
            key = self.widget.verticalHeaderItem(row).text()
            value = self.widget.item(row, 0).obj
            yield key, value



from magicclass import magicclass, PopUpMode

def _make_class(mode: PopUpMode, widget_type="none"):
    @magicclass(popup_mode=mode, widget_type=widget_type)
    class A:
        @magicclass(popup_mode=mode)
        class B:
            def b1(self, a: int): ...
            def b2(self, a: int): ...

        def a1(self, a: int): ...

        @B.wraps
        def b1(self, a: int): ...

    return A

def test_all_works():
    for mode in PopUpMode._member_names_:
        if mode == "dock":
            ui = _make_class(mode, "mainwindow")()
        elif mode == "dialog":
            # TODO: how to test dialog?
            continue
        else:
            ui = _make_class(mode)()
        ui.show(run=False)
        ui[1].changed()
        ui.B[0].changed()
        ui.B[1].changed()
        ui.close()


def _make_class_2(mode: PopUpMode):
    @magicclass(popup_mode=mode)
    class A:
        @magicclass(popup_mode=mode)
        class B:
            def b1(self, a: int): ...
            def b2(self, a: int): ...
            def b3(self, a: int): ...
            def b4(self, a: int): ...
            def b5(self, a: int): ...
            def b6(self, a: int): ...

        def a1(self, a: int): ...
        def a2(self, a: int): ...
        def a3(self, a: int): ...
        def a4(self, a: int): ...
        def a5(self, a: int): ...
        def a6(self, a: int): ...

        @B.wraps
        def b3(self, a: int): ...

    return A

def test_first():
    ui = _make_class_2(PopUpMode.first)()
    l = len(ui)

    ui["a2"].changed()
    assert len(ui) == l + 1
    assert ui[0] is ui["a2"].mgui

    ui["a5"].changed()
    assert len(ui) == l + 2
    assert ui[0] is ui["a5"].mgui


def test_last():
    ui = _make_class_2(PopUpMode.last)()
    l = len(ui)

    ui["a2"].changed()
    assert len(ui) == l + 1
    assert ui[-1] is ui["a2"].mgui

    ui["a5"].changed()
    assert len(ui) == l + 2
    assert ui[-1] is ui["a5"].mgui



from enum import Enum


class StringEnum(Enum):
    def __eq__(self, other):
        if isinstance(other, str):
            return self.name == other
        return super().__eq__(other)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}.{self.name}"

    def __str__(self) -> str:
        return self.name


class Rendering(StringEnum):
    """Volume rendering mode suppored in vtk."""

    composite = 0
    mip = 1
    minip = 2
    average = 3
    additive = 4


class Mode(StringEnum):
    volume = "volume"
    iso = "iso"
    lego = "lego"
    mesh = "mesh"
    wireframe = "wireframe"


class Representation(StringEnum):
    points = 0
    wireframe = 1
    surface = 2

    def __int__(self):
        return self.value


class AxesMode(StringEnum):
    none = 0
    wall = 1
    cartesian = 2
    cartesian_pos = 3
    triad = 4
    cube = 5
    corner = 6
    ruler = 7
    ruler_axes = 8
    box = 9
    circle = 10
    grid = 11
    polar = 12
    ruler1d = 13



from __future__ import annotations
from typing import TYPE_CHECKING
from qtpy import QtWidgets as QtW
from qtpy.QtCore import Qt
from napari.components.viewer_model import ViewerModel
from napari.qt import QtViewer

from ...widgets import FreeWidget

if TYPE_CHECKING:
    from napari import layers as nl


class QtLayerWidget(QtW.QSplitter):
    def __init__(self, parent: QtViewerWidget):
        super().__init__(Qt.Orientation.Vertical, parent)
        layercontrol = parent._qt_viewer.dockLayerControls.widget()
        self.addWidget(layercontrol)
        layerlist = parent._qt_viewer.dockLayerList.widget()
        self.addWidget(layerlist)


class QtViewerWidget(QtW.QSplitter):
    def __init__(self):
        super().__init__()
        self._viewer_model = ViewerModel(title="Model")
        self._qt_viewer = QtViewer(self._viewer_model)
        self._layer_widget = QtLayerWidget(self)
        self.addWidget(self._layer_widget)
        self.addWidget(self._qt_viewer)


class ViewerWidget(FreeWidget):
    def __init__(self):
        super().__init__()
        self._qtwidget = QtViewerWidget()
        self.set_widget(self._qtwidget)

    @property
    def layers(self):
        return self._qtwidget._viewer_model.layers

    @property
    def dims(self):
        return self._qtwidget._viewer_model.dims

    @property
    def axes(self):
        return self._qtwidget._viewer_model.axes

    @property
    def camera(self):
        return self._qtwidget._viewer_model.camera

    @property
    def cursor(self):
        return self._qtwidget._viewer_model.cursor

    @property
    def grid(self):
        return self._qtwidget._viewer_model.grid

    @property
    def scale_bar(self):
        return self._qtwidget._viewer_model.scale_bar

    @property
    def text_overlay(self):
        return self._qtwidget._viewer_model.text_overlay

    @property
    def cursor(self):
        return self._qtwidget._viewer_model.cursor

    def add_image(self, *args, **kwargs) -> nl.Image:
        return self._qtwidget._viewer_model.add_image(*args, **kwargs)

    def add_shapes(self, *args, **kwargs) -> nl.Shapes:
        return self._qtwidget._viewer_model.add_shapes(*args, **kwargs)

    def add_points(self, *args, **kwargs) -> nl.Points:
        return self._qtwidget._viewer_model.add_points(*args, **kwargs)

    def add_vectors(self, *args, **kwargs) -> nl.Vectors:
        return self._qtwidget._viewer_model.add_vectors(*args, **kwargs)

    def add_labels(self, *args, **kwargs) -> nl.Labels:
        return self._qtwidget._viewer_model.add_labels(*args, **kwargs)

    def add_surface(self, *args, **kwargs) -> nl.Surface:
        return self._qtwidget._viewer_model.add_surface(*args, **kwargs)

    def add_tracks(self, *args, **kwargs) -> nl.Tracks:
        return self._qtwidget._viewer_model.add_tracks(*args, **kwargs)



from magicclass import magicclass, MagicTemplate
from magicgui import magicgui, widgets

def test_magicgui():
    @magicclass
    class A(MagicTemplate):
        @magicgui
        def f(self, a: int, b: str = "x"):
            self.a = a

    ui = A()
    assert type(ui[0]) is widgets.FunctionGui
    assert ui[0].a.value == 0
    assert ui[0].b.value == "x"

    # macro should be recorded after called
    ui[0].a.value = 10
    assert len(ui.macro) == 1
    ui[0]()
    assert len(ui.macro) == 2
    assert str(ui.macro[-1]) == "ui.f(a=10, b='x')"
    assert hasattr(ui, "a")
    ui[0].a.value = 20
    ui[0]()
    assert len(ui.macro) == 3
    assert str(ui.macro[-1]) == "ui.f(a=20, b='x')"
    assert str(ui.macro[-2]) == "ui.f(a=10, b='x')"


def test_autocall_macro():
    """Auto-called macro should be recorded once."""
    @magicclass
    class A(MagicTemplate):
        @magicgui(auto_call=True)
        def f(self, a: int, b: str = "x"):
            self.a = a

    ui = A()
    ui[0].a.value = 1
    assert str(ui.macro[-1]) == "ui.f(a=1, b='x')"
    ui[0].a.value = 2
    assert str(ui.macro[-1]) == "ui.f(a=2, b='x')"

def test_wraps():
    @magicclass
    class A(MagicTemplate):
        @magicclass
        class B:
            def f1(self): ...
            def f2(self): ...
            def f3(self): ...

        @B.wraps
        @magicgui
        def f2(self, a: int, b: str = "x"):
            self.a = a

    ui = A()
    assert len(ui.B) == 3
    assert type(ui.B[1]) is widgets.FunctionGui
    assert ui.B[1].a.value == 0
    assert ui.B[1].b.value == "x"

    # macro should be recorded after called
    ui.B[1].a.value = 10
    assert len(ui.macro) == 1
    ui.B[1]()
    assert len(ui.macro) == 2
    assert str(ui.macro[-1]) == "ui.f2(a=10, b='x')"
    assert hasattr(ui, "a")
    ui.B[1].a.value = 20
    ui.B[1]()
    assert len(ui.macro) == 3
    assert str(ui.macro[-1]) == "ui.f2(a=20, b='x')"
    assert str(ui.macro[-2]) == "ui.f2(a=10, b='x')"

# NOTE: need fix on magicgui side
# def test_multi_gui():
#     @magicclass
#     class A:
#         @magicgui
#         def f(self, x: int):
#             return id(self)

#         @magicgui
#         def g(self, x: float):
#             return -id(self)

#     ui0 = A()
#     ui1 = A()
#     assert ui0["f"] is not ui1["f"]
#     assert ui0["g"] is not ui1["g"]
#     assert ui0["f"](0) == -ui0["g"](0)
#     assert ui1["f"](0) == -ui1["g"](0)
#     assert ui0["f"](0) != ui1["f"](0)



from enum import Enum


class Modifier(str, Enum):
    shift = "shift"
    control = "control"
    alt = "alt"


class Button(str, Enum):
    left = "left"
    right = "right"
    middle = "middle"


class Anchor(str, Enum):
    ...



from __future__ import annotations
import pyqtgraph as pg
from pyqtgraph import colormap as cmap
from typing import Generic, Iterator, Sequence, TypeVar, overload, MutableSequence
import numpy as np

from .components import Legend, Region, ScaleBar, TextItem
from .graph_items import (
    BarPlot,
    Curve,
    FillBetween,
    InfLine,
    LayerItem,
    Scatter,
    Histogram,
    TextGroup,
)
from .mouse_event import MouseClickEvent
from .._shared_utils import convert_color_code, to_rgba
from .._doc import write_docs
from ...widgets.utils import FreeWidget
from ..._app import get_app

BOTTOM = "bottom"
LEFT = "left"


class LayerList(MutableSequence[LayerItem]):
    """A napari-like layer list for plot item handling."""

    def __init__(self, parent: HasDataItems):
        self.parent = parent

    def __repr__(self) -> str:
        if len(self) == 0:
            return f"{type(self).__name__}()"
        s = ",\n\t".join(repr(layer) for layer in self)
        return f"{type(self).__name__}(\n\t{s}\n)"

    def __getitem__(self, key: int | str) -> LayerItem:
        if isinstance(key, int):
            return self.parent._items[key]
        elif isinstance(key, str):
            for item in self.parent._items:
                if item.name == key:
                    return item
            else:
                raise ValueError(f"Item '{key}' not found.")
        else:
            raise TypeError(f"Cannot use type {type(key)} as a key.")

    def __setitem__(self, key, value):
        raise NotImplementedError("Can't set item")

    def __delitem__(self, key: int | str):
        return self.parent._remove_item(key)

    def append(self, item: LayerItem):
        if not isinstance(item, LayerItem):
            raise TypeError(f"Cannot append type {type(item)}.")
        self.parent._add_item(item)

    def insert(self, pos: int, item: LayerItem):
        if not isinstance(item, LayerItem):
            raise TypeError(f"Cannot insert type {type(item)}.")
        self.parent._insert_item(pos, item)

    def __len__(self):
        return len(self.parent._items)

    def clear(self):
        for _ in range(len(self)):
            self.parent._remove_item(-1)

    def swap(self, pos0: int, pos1: int):
        return self.parent._swap_items(pos0, pos1)

    def move(self, source: int, destination: int):
        return self.parent._move_item(source, destination)


class HasDataItems:
    _items: list[LayerItem]

    @property
    def _graphics(self) -> pg.GraphicsWidget:
        """Target widget to add graphics items."""
        raise NotImplementedError()

    @property
    def layers(self) -> LayerList:
        return LayerList(self)

    @overload
    def add_curve(self, x: Sequence[float], **kwargs):
        ...

    @overload
    def add_curve(self, x: Sequence[float], y: Sequence[float], **kwargs):
        ...

    @write_docs
    def add_curve(
        self,
        x=None,
        y=None,
        face_color=None,
        edge_color=None,
        color=None,
        size: float = 7,
        name: str | None = None,
        lw: float = 1,
        ls: str = "-",
        symbol=None,
    ):
        """
        Add a line plot like ``plt.plot(x, y)``.

        Parameters
        ----------
        {x}
        {y}
        {face_color}
        {edge_color}
        {color}
        size: float, default is 7
            Symbol size.
        {name}
        {lw}
        {ls}
        {symbol}

        Returns
        -------
        Curve
            A plot item of a curve.
        """
        x, y = _check_xy(x, y)
        name = self._find_unique_name(name or "Curve")
        face_color, edge_color = _check_colors(face_color, edge_color, color)
        item = Curve(
            x,
            y,
            face_color=face_color,
            edge_color=edge_color,
            size=size,
            name=name,
            lw=lw,
            ls=ls,
            symbol=symbol,
        )
        self._add_item(item)
        return item

    @overload
    def add_scatter(self, x: Sequence[float], **kwargs):
        ...

    @overload
    def add_scatter(self, x: Sequence[float], y: Sequence[float], **kwargs):
        ...

    @write_docs
    def add_scatter(
        self,
        x=None,
        y=None,
        face_color=None,
        edge_color=None,
        color=None,
        size: float = 7,
        name: str | None = None,
        lw: float = 1,
        ls: str = "-",
        symbol="o",
    ):
        """
        Add scatter plot like ``plt.scatter(x, y)``.

        Parameters
        ----------
        {x}
        {y}
        {face_color}
        {edge_color}
        {color}
        size: float, default is 7
            Symbol size.
        {name}
        {lw}
        {ls}
        {symbol}

        Returns
        -------
        Scatter
            A plot item of the scatter plot.
        """
        x, y = _check_xy(x, y)
        name = self._find_unique_name(name or "Scatter")
        face_color, edge_color = _check_colors(face_color, edge_color, color)
        item = Scatter(
            x,
            y,
            face_color=face_color,
            edge_color=edge_color,
            size=size,
            name=name,
            lw=lw,
            ls=ls,
            symbol=symbol,
        )
        self._add_item(item)
        return item

    @write_docs
    def add_hist(
        self,
        data: Sequence[float],
        bins: int | Sequence | str = 10,
        range=None,
        density: bool = False,
        face_color=None,
        edge_color=None,
        color=None,
        name: str | None = None,
        lw: float = 1,
        ls: str = "-",
    ):
        """
        Add histogram like ``plt.hist(data)``.

        Parameters
        ----------
        data : array-like
            Data for histogram constrction.
        bins : int, sequence of float or str, default is 10
            Bin numbers. See ``np.histogram`` for detail.
        range : two floats, optional
            Bin ranges. See ``np.histogram`` for detail.
        density : bool, default is False
            If true, plot the density instead of the counts. See ``np.histogram`` for
            detail.
        {face_color}
        {edge_color}
        {color}
        {name}
        {lw}
        {ls}

        Returns
        -------
        Histogram
            A plot item of the histogram.
        """
        name = self._find_unique_name(name or "Histogram")
        face_color, edge_color = _check_colors(face_color, edge_color, color)
        item = Histogram(
            data,
            bins=bins,
            range=range,
            density=density,
            face_color=face_color,
            edge_color=edge_color,
            name=name,
            lw=lw,
            ls=ls,
        )
        self._add_item(item)
        return item

    @overload
    def add_bar(self, x: Sequence[float], **kwargs):
        ...

    @overload
    def add_bar(self, x: Sequence[float], y: Sequence[float], **kwargs):
        ...

    @write_docs
    def add_bar(
        self,
        x=None,
        y=None,
        width: float = 0.6,
        face_color=None,
        edge_color=None,
        color=None,
        name: str | None = None,
        lw: float = 1,
        ls: str = "-",
    ):
        """
        Add a bar plot like ``plt.bar(x, y)``.

        Parameters
        ----------
        {x}
        {y}
        width : float, default is 0.6
            Width of each bar.
        {face_color}
        {edge_color}
        {color}
        {name}
        {lw}
        {ls}

        Returns
        -------
        BarPlot
            A plot item of the bar plot.
        """
        x, y = _check_xy(x, y)
        name = self._find_unique_name(name or "Bar")
        face_color, edge_color = _check_colors(face_color, edge_color, color)
        item = BarPlot(
            x,
            y,
            width=width,
            face_color=face_color,
            edge_color=edge_color,
            name=name,
            lw=lw,
            ls=ls,
        )
        self._add_item(item)
        return item

    @overload
    def add_fillbetween(self, x: Sequence[float], **kwargs):
        ...

    @overload
    def add_fillbetween(self, x: Sequence[float], y: Sequence[float], **kwargs):
        ...

    @write_docs
    def add_fillbetween(
        self,
        x=None,
        y1=None,
        y2=None,
        face_color=None,
        edge_color=None,
        color=None,
        name: str | None = None,
        lw: float = 1,
        ls: str = "-",
    ):
        x, y1 = _check_xy(x, y1)
        name = self._find_unique_name(name or "FillBetween")
        face_color, edge_color = _check_colors(face_color, edge_color, color)

        item = FillBetween(
            x,
            y1,
            y2,
            face_color=face_color,
            edge_color=edge_color,
            name=name,
            lw=lw,
            ls=ls,
        )
        self._add_item(item)

    @overload
    def add_infline(
        self,
        slope: float,
        intercept: float,
        color=None,
        name: str | None = None,
        lw: float = 1,
        ls: str = "-",
    ):
        ...

    @overload
    def add_infline(
        self,
        pos: tuple[float, float],
        degree: float,
        color=None,
        name: str | None = None,
        lw: float = 1,
        ls: str = "-",
    ):
        ...

    def add_infline(
        self,
        *args,
        color=None,
        name: str | None = None,
        lw: float = 1,
        ls: str = "-",
        **kwargs,
    ):
        if kwargs:
            if args:
                raise TypeError(
                    "Cannot mix args and kwargs for infinite line parameters."
                )
            keys = set(kwargs.keys())
            if keys <= {"pos", "angle"}:
                args = (kwargs.get("pos", (0, 0)), kwargs.get("angle", 0))
            elif keys <= {"slope", "intercept"}:
                args = (kwargs.get("slope", (0, 0)), kwargs.get("intercept", 0))
            else:
                raise ValueError(f"{kwargs} is invalid input.")

        nargs = len(args)
        if nargs == 1:
            arg0 = args[0]
            if np.isscalar(arg0):
                angle = np.rad2deg(np.arctan(arg0))
                pos = (0, 0)
            else:
                pos = arg0
                angle = 90
        elif nargs == 2:
            arg0, arg1 = args
            if np.isscalar(arg0):
                angle = np.rad2deg(np.arctan(arg0))
                pos = (0, arg1)
            else:
                pos = arg0
                angle = arg1
        else:
            raise TypeError(
                "Arguments of 'add_infline' should be either 'add_infline(slope, intercept)' "
                "or 'add_infline(pos, degree)'."
            )

        item = InfLine(pos, angle, edge_color=color, name=name, lw=lw, ls=ls)
        self._add_item(item)

    @overload
    def add_text(self, x: float, y: float, text: str, **kwargs):
        ...

    @overload
    def add_text(
        self, x: Sequence[float], y: Sequence[float], text: Sequence[str], **kwargs
    ):
        ...

    def add_text(self, x, y, text, color=None, name=None):
        if np.isscalar(x) and np.isscalar(y):
            x = [x]
            y = [y]
            text = [text]
        item = TextGroup(x, y, text, color, name)
        self._add_item(item)

    def _add_item(self, item: LayerItem):
        item.zorder = len(self._items)
        self._graphics.addItem(item.native)
        self._items.append(item)

    def _insert_item(self, pos: int, item: LayerItem):
        self._graphics.addItem(item.native)
        self._items.insert(pos, item)
        self._reorder()

    def _swap_items(self, pos0: int, pos1: int):
        item0 = self._items[pos0]
        item1 = self._items[pos1]
        self._items[pos0] = item1
        self._items[pos1] = item0
        self._reorder()

    def _move_item(self, source: int, destination: int):
        if source < destination:
            destination -= 1
        item = self._items.pop(source)
        self._items.insert(destination, item)
        self._reorder()

    def _remove_item(self, item: LayerItem | int | str):
        if isinstance(item, LayerItem):
            i = self._items.index(item)
        elif isinstance(item, int):
            if item < 0:
                item += len(self._items)
            i = item
        elif isinstance(item, str):
            for i, each in enumerate(self._items):
                if each.name == item:
                    break
            else:
                raise ValueError(f"No item named {item}")

        if i < 0:
            raise ValueError(f"Item {item} not found")
        item = self._items.pop(i)
        self._graphics.removeItem(item.native)

    def _reorder(self):
        for i, item in enumerate(self._items):
            item.zorder = i
        return None

    def _find_unique_name(self, prefix: str):
        existing_names = [item.name for item in self._items]
        name = prefix
        i = 0
        while name in existing_names:
            name = f"{prefix}-{i}"
            i += 1
        return name


class HasViewBox(HasDataItems):
    def __init__(self, viewbox: pg.ViewBox):
        self._viewbox = viewbox
        self._items: list[LayerItem] = []

        # prepare mouse event
        self.mouse_click_callbacks = []

        # This ROI is not editable. Mouse click event will use it to determine
        # the origin of the coordinate system.
        self._coordinate_fiducial = pg.ROI((0, 0))
        self._coordinate_fiducial.setVisible(False)
        self._viewbox.addItem(self._coordinate_fiducial, ignoreBounds=True)

        self._enabled = True

    def _mouse_clicked(self, e):
        # NOTE: Mouse click event needs a reference item to map coordinates.
        # Here plot items always have a linear region item as a default one,
        # we can use it as the referene.
        ev = MouseClickEvent(e, self._coordinate_fiducial)
        x, y = ev.pos()
        [xmin, xmax], [ymin, ymax] = self._viewbox.viewRange()

        if xmin <= x <= xmax and ymin <= y <= ymax:
            for callback in self.mouse_click_callbacks:
                callback(ev)

    @property
    def xlim(self):
        """Range limits of X-axis."""
        (xmin, xmax), _ = self._viewbox.viewRange()
        return xmin, xmax

    @xlim.setter
    def xlim(self, value: tuple[float, float]):
        self._viewbox.setXRange(*value)

    @property
    def ylim(self):
        """Range limits of Y-axis."""
        _, (ymin, ymax) = self._viewbox.viewRange()
        return ymin, ymax

    @ylim.setter
    def ylim(self, value: tuple[float, float]):
        self._viewbox.setYRange(*value)

    @property
    def enabled(self) -> bool:
        """Mouse interactivity"""
        return self._enabled

    @enabled.setter
    def enabled(self, value: bool):
        self._viewbox.setMouseEnabled(value, value)
        self._enabled = value

    interactive = enabled

    def _update_scene(self):
        raise NotImplementedError()

    @property
    def border(self):
        return to_rgba(self._viewbox.border)

    @border.setter
    def border(self, value):
        value = convert_color_code(value)
        self._viewbox.setBorder(value)


class SimpleViewBox(HasViewBox):
    def __init__(self):
        super().__init__(pg.ViewBox())

    @property
    def _graphics(self):
        return self._viewbox


class PlotItem(HasViewBox):
    """
    A 1-D plot item that has similar API as napari Viewer.
    """

    def __init__(self, viewbox: pg.ViewBox | None = None):
        if viewbox is None:
            viewbox = pg.ViewBox()

        self.pgitem = pg.PlotItem(viewBox=viewbox)
        super().__init__(self.pgitem.vb)

        # prepare region item
        self._region = Region()
        self._region.visible = False
        self.pgitem.addItem(self._region.native, ignoreBounds=True)

        # prepare legend item
        self._legend = Legend()
        self._legend.native.setParentItem(self._viewbox)
        self.pgitem.legend = self._legend.native

        # initialize private attributes
        self._xlabel = ""
        self._ylabel = ""

    @property
    def _graphics(self):
        return self.pgitem

    @property
    def region(self) -> Region:
        """Linear region item."""
        return self._region

    @property
    def legend(self) -> Legend:
        """Legend item."""
        return self._legend

    @property
    def xlabel(self) -> str:
        """Label of X-axis."""
        return self._xlabel

    @xlabel.setter
    def xlabel(self, label: str):
        self.pgitem.setLabel(BOTTOM, label)
        self._xlabel = label

    @property
    def ylabel(self) -> str:
        """Label of Y-axis."""
        return self._ylabel

    @ylabel.setter
    def ylabel(self, label: str):
        self.pgitem.setLabel(LEFT, label)
        self._ylabel = label

    @property
    def title(self) -> str:
        """Title text of the graph."""
        return self.pgitem.titleLabel.text

    @title.setter
    def title(self, value: str):
        value = str(value)
        self.pgitem.setTitle(value)

    def _update_scene(self):
        # Since plot item does not have graphics scene before being added to
        # a graphical layout, mouse event should be connected afterward.
        self.pgitem.scene().sigMouseClicked.connect(self._mouse_clicked)


class ViewBoxExt(pg.ViewBox):
    def __init__(
        self,
        parent=None,
        border=None,
        lockAspect=False,
        enableMouse=True,
        invertY=False,
        enableMenu=True,
        name=None,
        invertX=False,
        defaultPadding=0.02,
    ):
        pg.ViewBox.__init__(**locals())
        from pyqtgraph import icons

        self.button = pg.ButtonItem(icons.getGraphPixmap("ctrl"), 14, self)
        self.button.hide()

    def hoverEvent(self, ev):
        try:
            if ev.enter:
                self.button.show()
            if ev.exit:
                self.button.hide()
        except RuntimeError:
            pass


class ImageItem(HasViewBox):
    def __init__(
        self, viewbox: pg.ViewBox | None = None, lock_contrast_limits: bool = False
    ):
        if viewbox is None:
            viewbox = ViewBoxExt(lockAspect=True, invertY=True)
        self._lock_contrast_limits = lock_contrast_limits
        super().__init__(viewbox)
        self._image_item = pg.ImageItem()
        tr = self._image_item.transform().translate(-0.5, -0.5)
        self._image_item.setTransform(tr)

        self._viewbox.addItem(self._image_item)

        # prepare text overlay
        self._text_overlay = TextItem(text="", color="gray")
        self._text_overlay.native.setParentItem(self._viewbox)

        # prepare scale bar
        self._scale_bar = ScaleBar()
        self._scale_bar.visible = False
        self._scale_bar.native.setParentItem(self._viewbox)
        self._scale_bar.native.anchor((1, 1), (1, 1), offset=(-20, -20))

        # prepare title and labels
        self._title = TextItem(text="", color="white", anchor=(0.5, 1.1))
        self._title.pos = [1, 0]
        self._viewbox.addItem(self._title.native)
        self._title.visible = False

        self._xlabel = TextItem(text="", color="white", anchor=(0.5, -0.1))
        self._xlabel.pos = [1, 1]
        self._viewbox.addItem(self._xlabel.native)
        self._xlabel.visible = False

        self._ylabel = TextItem(text="", color="white", anchor=(0.5, 1.1), angle=90)
        self._ylabel.pos = [0, 1]
        self._viewbox.addItem(self._ylabel.native)
        self._ylabel.visible = False

        if isinstance(viewbox, ViewBoxExt):
            # prepare LUT histogram
            self._hist = pg.HistogramLUTItem(orientation="horizontal")
            self._hist.vb.setBackgroundColor([0, 0, 0, 0.2])
            self._hist.setParentItem(self._viewbox)
            self._hist.setVisible(False)

            @viewbox.button.clicked.connect
            def _(e):
                visible = not self._hist.isVisible()
                self._hist.setVisible(visible)
                if visible:
                    self._hist._updateView()
                    width = min(160, self._viewbox.width())
                    self._hist.setFixedWidth(width)

        self._cmap = "gray"

    def _update_scene(self):
        # Since plot item does not have graphics scene before being added to
        # a graphical layout, mouse event should be connected afterward.
        self._image_item.scene().sigMouseClicked.connect(self._mouse_clicked)

    @property
    def text_overlay(self) -> TextItem:
        """Text overlay on the image."""
        return self._text_overlay

    @property
    def title(self) -> str:
        """Title text of the graph."""
        return self._title.text

    @title.setter
    def title(self, value: str):
        self._title.text = value

    @property
    def xlabel(self) -> str:
        """Label of X-axis."""
        return self._xlabel.text

    @xlabel.setter
    def xlabel(self, value: str):
        self._xlabel.text = value

    @property
    def ylabel(self) -> str:
        """Label of Y-axis."""
        return self._ylabel.text

    @xlabel.setter
    def ylabel(self, value: str):
        self._ylabel.text = value

    @property
    def scale_bar(self) -> ScaleBar:
        """Scale bar on the image."""
        return self._scale_bar

    @property
    def lock_contrast_limits(self):
        """True if enable auto-contrast."""
        return self._lock_contrast_limits

    @lock_contrast_limits.setter
    def lock_contrast_limits(self, value: bool):
        self._lock_contrast_limits = bool(value)

    @property
    def _graphics(self):
        return self._viewbox

    @property
    def image(self) -> np.ndarray | None:
        """Image data"""
        if self._image_item.image is None:
            return None
        else:
            return self._image_item.image.T

    @image.setter
    def image(self, image: np.ndarray):
        no_image = self._image_item.image is None
        if no_image:
            auto_levels = True
        else:
            auto_levels = not self._lock_contrast_limits
            clims = self.contrast_limits
        img = np.asarray(image)
        self._image_item.setImage(img.T, autoLevels=auto_levels)
        self._hist.setImageItem(self._image_item)
        self._hist._updateView()
        if no_image:
            self._viewbox.autoRange()
        if not auto_levels:
            self.contrast_limits = clims
        sy, sx = img.shape[-2:]
        self._title.pos = [sx / 2, self._title.pos[1]]
        self._title.visible = True
        self._xlabel.pos = [sx / 2, sy]
        self._xlabel.visible = True
        self._ylabel.pos = [0, sy / 2]
        self._ylabel.visible = True

    @image.deleter
    def image(self):
        self._image_item.clear()
        self._title.visible = False
        self._xlabel.visible = False
        self._ylabel.visible = False

    @property
    def contrast_limits(self) -> list[float, float]:
        """Contrast limits of image"""
        return self._image_item.levels

    @contrast_limits.setter
    def contrast_limits(self, value: tuple[float, float]):
        self._hist.setLevels(*value)

    @property
    def cmap(self):
        """Color map"""
        return self._cmap

    @cmap.setter
    def cmap(self, value):
        if isinstance(value, str):
            _cmap = cmap.get(value, source="matplotlib")
        else:
            _cmap = value
        self._hist.gradient.setColorMap(_cmap)
        self._cmap = value


class HasBackground(FreeWidget):
    @property
    def background_color(self):
        return self.layoutwidget._background

    @background_color.setter
    def background_color(self, value):
        value = convert_color_code(value)
        self.layoutwidget.setBackground(value)


class QtPlotCanvas(HasBackground, PlotItem):
    """A 1-D data viewer that have similar API as napari Viewer."""

    def __init__(self, **kwargs):
        app = get_app()
        # prepare widget
        PlotItem.__init__(self)
        self.layoutwidget = pg.GraphicsLayoutWidget()
        self.layoutwidget.addItem(self.pgitem)
        self._update_scene()

        super().__init__(**kwargs)
        self.set_widget(self.layoutwidget)


class QtImageCanvas(HasBackground, ImageItem):
    """A 2-D image viewer that have similar API as napari Viewer."""

    def __init__(self, lock_contrast_limits: bool = False, **kwargs):
        app = get_app()
        # prepare widget
        ImageItem.__init__(self, lock_contrast_limits=lock_contrast_limits)
        self.layoutwidget = pg.GraphicsLayoutWidget()
        self.layoutwidget.addItem(self._viewbox)
        self._update_scene()

        super().__init__(**kwargs)
        self.set_widget(self.layoutwidget)

    @property
    def background_color(self):
        return self.layoutwidget._background

    @background_color.setter
    def background_color(self, value):
        value = convert_color_code(value)
        self.layoutwidget.setBackground(value)


class Qt2YPlotCanvas(HasBackground):
    """A plot canvas with two Y-axis."""

    def __init__(self, **kwargs):
        app = get_app()
        self.layoutwidget = pg.GraphicsLayoutWidget()

        item_l = PlotItem()
        item_r = SimpleViewBox()

        self.layoutwidget.addItem(item_l.pgitem)

        item_l.pgitem.scene().addItem(item_r._viewbox)
        item_l.pgitem.getAxis("right").linkToView(item_r._viewbox)
        item_r._viewbox.setXLink(item_l.pgitem)

        item_l._update_scene()
        item_l.pgitem.showAxis("right")

        self._plot_items: tuple[PlotItem, SimpleViewBox] = (item_l, item_r)

        self.updateViews()
        item_l.pgitem.vb.sigResized.connect(self.updateViews)

        super().__init__(**kwargs)
        self.set_widget(self.layoutwidget)

    def __getitem__(self, k: int) -> PlotItem:
        return self._plot_items[k]

    def updateViews(self):
        item_l, item_r = self._plot_items
        item_r._viewbox.setGeometry(item_l._viewbox.sceneBoundingRect())
        item_r._viewbox.linkedViewChanged(item_l._viewbox, item_r._viewbox.XAxis)


_C = TypeVar("_C", bound=HasViewBox)


class _MultiPlot(HasBackground, Generic[_C]):
    _base_item_class: type[_C]

    def __init__(
        self,
        nrows: int = 1,
        ncols: int = 1,
        sharex: bool = False,
        sharey: bool = False,
        **kwargs,
    ):
        """
        Multi-axes ``pyqtgraph`` canvas widget. Can contain multiple objects
        of {cls}.

        Parameters
        ----------
        nrows : int, default is 1
            Initial rows of axes.
        ncols : int, default is 1
            Initail columns of axes.
        sharex : bool, default is False
            If true, all the x-axes will be linked.
        sharey : bool, default is False
            If true, all the y-axes will be linked.
        """
        app = get_app()
        self.layoutwidget = pg.GraphicsLayoutWidget()
        self._axes: list[_C] = []
        self._sharex = sharex
        self._sharey = sharey

        super().__init__(**kwargs)
        self.set_widget(self.layoutwidget)

        if nrows * ncols > 0:
            for r in range(nrows):
                for c in range(ncols):
                    self.addaxis(r, c)

    def __init_subclass__(cls) -> None:
        """Update doc."""
        init = cls.__init__
        init.__doc__ = init.__doc__.format(cls=cls._base_item_class.__name__)

    def addaxis(
        self,
        row: int | None = None,
        col: int | None = None,
        rowspan: int = 1,
        colspan: int = 1,
    ) -> _C:
        """Add a new axis to widget."""
        item = self._base_item_class()
        self._axes.append(item)
        self.layoutwidget.addItem(item._graphics, row, col, rowspan, colspan)
        item._update_scene()
        if self._sharex and len(self._axes) > 1:
            item._viewbox.setXLink(self._axes[0]._viewbox)
        if self._sharey and len(self._axes) > 1:
            item._viewbox.setYLink(self._axes[0]._viewbox)
        return item

    def __getitem__(self, k: int) -> _C:
        return self._axes[k]

    def __delitem__(self, k: int):
        item = self._axes[k]
        self.layoutwidget.removeItem(item._graphics)

    def __iter__(self) -> Iterator[_C]:
        return iter(self._axes)


class QtMultiPlotCanvas(_MultiPlot[PlotItem]):
    """A pyqtgraph-based canvas with multiple plot."""

    _base_item_class = PlotItem


class QtMultiImageCanvas(_MultiPlot[ImageItem]):
    """A pyqtgraph-based canvas with multiple images."""

    _base_item_class = ImageItem


def _check_xy(x, y):
    if y is None:
        if x is None:
            x = []
            y = []
        else:
            y = x
            x = np.arange(len(y))

    return x, y


def _check_colors(face_color, edge_color, color):
    if color is None:
        return face_color, edge_color
    else:
        if face_color is None and edge_color is None:
            return color, color
        else:
            raise ValueError(
                "Cannot set 'color' and either 'face_color' or "
                "'edge_color' at the same time."
            )



from __future__ import annotations
from typing import TYPE_CHECKING
import weakref
from qtpy.QtWidgets import QWidget, QVBoxLayout, QGridLayout, QHBoxLayout
from magicgui.widgets import Widget
from magicgui.widgets._concrete import merge_super_sigs as _merge_super_sigs
from magicgui.backends._qtpy.widgets import QBaseWidget

if TYPE_CHECKING:
    from .._gui import BaseGui, ContextMenuGui


class _NotInitialized:
    """This class helps better error handling."""

    def __init__(self, msg: str):
        self.msg = msg

    def __getattr__(self, key: str):
        raise RuntimeError(self.msg)


class FreeWidget(Widget):
    """A Widget class with any QWidget as a child."""

    _widget = _NotInitialized(
        "Widget is not correctly initialized. Must call `super().__init__` before using "
        "the widget."
    )

    def __init__(self, layout="vertical", **kwargs):
        super().__init__(
            widget_type=QBaseWidget, backend_kwargs={"qwidg": QWidget}, **kwargs
        )
        self.native: QWidget
        self.central_widget: QWidget | None = None
        if layout == "vertical":
            self.native.setLayout(QVBoxLayout())
        elif layout == "horizontal":
            self.native.setLayout(QHBoxLayout())
        elif layout == "grid":
            self.native.setLayout(QGridLayout())
        else:
            raise ValueError(layout)
        self.native.setContentsMargins(0, 0, 0, 0)
        self._magicclass_parent_ref = None

    def set_widget(self, widget: QWidget, *args):
        """Set the central widget to the widget."""
        self.native.layout().addWidget(widget, *args)
        widget.setParent(self.native)
        self.central_widget = widget

    def remove_widget(self, widget: QWidget):
        """Set the central widget from the widget."""
        self.native.layout().removeWidget(widget)
        widget.setParent(None)
        self.central_widget = None

    def set_contextmenu(self, contextmenugui: ContextMenuGui):
        from .._gui import ContextMenuGui

        if not isinstance(contextmenugui, ContextMenuGui):
            raise TypeError
        from .._gui.utils import set_context_menu

        set_context_menu(contextmenugui, self)

    @property
    def __magicclass_parent__(self) -> BaseGui | None:
        """Return parent magic class if exists."""
        if self._magicclass_parent_ref is None:
            return None
        parent = self._magicclass_parent_ref()
        return parent

    @__magicclass_parent__.setter
    def __magicclass_parent__(self, parent) -> None:
        if parent is None:
            return
        self._magicclass_parent_ref = weakref.ref(parent)


def merge_super_sigs(cls):
    cls = _merge_super_sigs(cls)
    cls.__module__ = "magicclass.widgets"
    return cls



from magicclass import magicclass, field, MagicTemplate
from unittest.mock import MagicMock

def test_predefinition():
    class Base(MagicTemplate):
        a = field(int)
        def shared_func_1(self, a: int):
            self.a.value = a
        def abstract_func(self): pass
        def shared_func_2(self, a: int):
            self.a.value = a

    @magicclass
    class A(Base):
        a = field(int)  # field needs re-definition
        def abstract_func(self): pass
        def func_a(self): pass

    @magicclass
    class B(Base):
        def func_b(self): pass
        a = field(int)
        def abstract_func(self): pass

    a = A()
    b = B()

    assert a[0].name == "a"
    assert a[1].name == "shared_func_1"
    assert a[2].name == "abstract_func"
    assert a[3].name == "shared_func_2"
    assert a[4].name == "func_a"

    assert b[0].name == "a"
    assert b[1].name == "shared_func_1"
    assert b[2].name == "abstract_func"
    assert b[3].name == "shared_func_2"
    assert b[4].name == "func_b"

    # test value change

    a.shared_func_1(10)
    assert a.a.value == 10
    assert b.a.value == 0

    b.shared_func_2(15)
    assert a.a.value == 10
    assert b.a.value == 15


def test_field():
    mock = MagicMock()

    class Base(MagicTemplate):
        a = field(int)
        @a.connect
        def _callback(self):
            mock()

    @magicclass
    class A(Base):
        pass

    @magicclass
    class B(Base):
        pass

    a = A()
    b = B()

    assert len(a) == 1
    assert len(b) == 1
    mock.assert_not_called()
    a.a.value = 1
    assert b.a.value == 0
    mock.assert_called_once()
    mock.reset_mock()
    b.a.value = 2
    assert a.a.value == 1
    mock.assert_called_once()

def test_nested_class():
    class Base(MagicTemplate):
        result = field(str)
        @magicclass
        class X(MagicTemplate):
            def func(self): ...
        @X.wraps
        def func(self):
            self.result.value = self.__class__.__name__

    @magicclass
    class A(Base):
        pass

    @magicclass
    class B(Base):
        pass

    a = A()
    b = B()

    assert a["X"] is not b["X"]

    a.X["func"].changed()
    assert a.result.value == "A"
    assert b.result.value == ""
    b.X["func"].changed()
    assert a.result.value == "A"
    assert b.result.value == "B"



"""
Jupyter QtConsole widget with callback signals.

You can embed a console widget in a magic class.

.. code-block:: python

    from magicclass import magicclass
    from magicclass.widgets import QtConsole

    class Main:
        console = QtConsole()

    ui = Main()
    ui.show()

There are some additional methods that would be very useful for developing a Python script
executable GUI.

.. code-block:: python

    # programmatically add code to console.
    ui.console.value = "a = 1"

    # callback when code is executed.
    @console.executed.connect
    def _():
        print("executed!")

    # programmatically execute code.
    ui.console.execute()

"""

from .widgets import QtConsole

__all__ = ["QtConsole"]



from __future__ import annotations
import numpy as np
from numpy.typing import ArrayLike
from vispy import scene

from .layer3d import Image, IsoSurface, Surface
from .layerlist import LayerList
from ...widgets import FreeWidget
from ...types import Color


class Vispy3DCanvas(FreeWidget):
    """
    A Vispy canvas for 3-D object visualization.

    Very similar to napari. This widget can be used independent of napari, or
    as a mini-viewer of napari.
    """

    def __init__(self):
        super().__init__()
        self._scene = scene.SceneCanvas()
        grid = self._scene.central_widget.add_grid()
        self._viewbox = grid.add_view()
        self._viewbox.camera = scene.ArcballCamera(fov=0)
        self._layerlist = LayerList()

        self._scene.create_native()
        self.set_widget(self._scene.native)

    @property
    def layers(self):
        """Return the layer list."""
        return self._layerlist

    @property
    def camera(self):
        """Return the native camera."""
        return self._viewbox.camera

    def add_image(
        self,
        data: ArrayLike,
        *,
        contrast_limits: tuple[float, float] = None,
        rendering: str = "mip",
        iso_threshold: float | None = None,
        attenuation: float = 1.0,
        cmap: str = "grays",
        gamma: float = 1.0,
        interpolation: str = "linear",
    ):
        image = Image(
            data,
            self._viewbox,
            contrast_limits=contrast_limits,
            rendering=rendering,
            iso_threshold=iso_threshold,
            attenuation=attenuation,
            cmap=cmap,
            gamma=gamma,
            interpolation=interpolation,
        )

        self._layerlist.append(image)
        self._viewbox.camera.scale_factor = max(data.shape)
        self._viewbox.camera.center = [s / 2 - 0.5 for s in data.shape]
        return image

    def add_isosurface(
        self,
        data: ArrayLike,
        *,
        contrast_limits: tuple[float, float] | None = None,
        iso_threshold: float | None = None,
        face_color: Color | None = None,
        edge_color: Color | None = None,
        shading: str = "smooth",
    ):
        surface = IsoSurface(
            data,
            self._viewbox,
            contrast_limits=contrast_limits,
            iso_threshold=iso_threshold,
            edge_color=edge_color,
            face_color=face_color,
            shading=shading,
        )

        self._layerlist.append(surface)
        self._viewbox.camera.scale_factor = max(data.shape)
        self._viewbox.camera.center = [s / 2 - 0.5 for s in data.shape]
        return surface

    def add_surface(
        self,
        data: tuple[ArrayLike, ArrayLike] | tuple[ArrayLike, ArrayLike, ArrayLike],
        *,
        face_color: Color | None = None,
        edge_color: Color | None = None,
        shading: str = "smooth",
    ):
        surface = Surface(
            data,
            self._viewbox,
            face_color=face_color,
            edge_color=edge_color,
            shading=shading,
        )
        self._layerlist.append(surface)
        mins = np.min(data[0], axis=0)
        maxs = np.max(data[0], axis=0)
        self._viewbox.camera.scale_factor = max(maxs - mins)
        self._viewbox.camera.center = [(s1 + s0) / 2 for s0, s1 in zip(mins, maxs)]
        return surface



from __future__ import annotations
from typing import Any, Callable, Sequence, TypeVar
import warnings
from qtpy.QtWidgets import QMenuBar, QWidget, QMainWindow, QBoxLayout, QDockWidget
from qtpy.QtCore import Qt
from magicgui.widgets import Container, MainWindow, Label, FunctionGui, Image, Table
from magicgui.widgets._bases import Widget, ButtonWidget, ValueWidget, ContainerWidget
from magicgui.widgets._concrete import _LabeledWidget, ContainerWidget
from macrokit import Symbol


from .keybinding import register_shortcut
from .mgui_ext import PushButtonPlus
from .toolbar import ToolBarGui, QtTabToolBar
from .menu_gui import MenuGui, ContextMenuGui
from ._base import (
    BaseGui,
    PopUpMode,
    ErrorMode,
    value_widget_callback,
    nested_function_gui_callback,
)
from .utils import (
    copy_class,
    format_error,
    set_context_menu,
)
from ..widgets import (
    ButtonContainer,
    GroupBoxContainer,
    FrameContainer,
    ListContainer,
    SubWindowsContainer,
    ScrollableContainer,
    DraggableContainer,
    CollapsibleContainer,
    HCollapsibleContainer,
    SplitterContainer,
    StackedContainer,
    TabbedContainer,
    ToolBoxContainer,
    FreeWidget,
)

from ..utils import iter_members, Tooltips
from ..fields import MagicField
from ..signature import get_additional_option
from .._app import run_app

# For Containers that belong to these classes, menubar must be set to _qwidget.layout().
_USE_OUTER_LAYOUT = (
    ScrollableContainer,
    DraggableContainer,
    SplitterContainer,
    TabbedContainer,
)

_MCLS_PAREMT = "__magicclass_parent__"


class ClassGuiBase(BaseGui):
    # This class is always inherited by @magicclass decorator.
    _component_class = PushButtonPlus
    _container_widget: type
    _remove_child_margins: bool
    native: QWidget

    def _create_widget_from_field(self, name: str, fld: MagicField):
        if fld.not_ready():
            raise TypeError(
                f"MagicField {name} does not contain enough information for widget creation"
            )

        fld.name = fld.name or name
        widget = fld.get_widget(self)

        if isinstance(widget, BaseGui):
            widget.__magicclass_parent__ = self
            self.__magicclass_children__.append(widget)
            widget._my_symbol = Symbol(name)

        if isinstance(widget, (ValueWidget, ContainerWidget)):
            # If the field has callbacks, connect it to the newly generated widget.
            if (
                isinstance(widget, ValueWidget) or hasattr(widget, "value")
            ) and fld.record:
                # By default, set value function will be connected to the widget.
                getvalue = type(fld) is MagicField
                f = value_widget_callback(self, widget, name, getvalue=getvalue)
                widget.changed.connect(f)

        elif fld.callbacks:
            warnings.warn(
                f"{type(widget).__name__} does not have value-change callback. "
                "Connecting callback functions does no effect.",
                UserWarning,
            )

        return widget

    def _convert_attributes_into_widgets(self):
        """
        This function is called in dynamically created __init__. Methods, fields and nested
        classes are converted to magicgui widgets.
        """
        cls = self.__class__

        # Add class docstring as tooltip.
        _tooltips = Tooltips(cls)
        self.tooltip = _tooltips.desc

        # Bind all the methods and annotations
        n_insert = 0
        base_members = {x[0] for x in iter_members(self._container_widget)}
        base_members |= {x[0] for x in iter_members(ClassGuiBase)}

        _hist: list[tuple[str, str, str]] = []  # for traceback
        _annot = ClassGuiBase.__annotations__.keys()
        _ignore_types = (property, classmethod, staticmethod)
        for name, attr in filter(lambda x: x[0] not in base_members, iter_members(cls)):
            if name in _annot or isinstance(attr, _ignore_types):
                continue

            try:
                if isinstance(attr, type):
                    # Nested magic-class
                    widget = attr()
                    object.__setattr__(self, name, widget)

                elif isinstance(attr, MagicField):
                    # If MagicField is given by field() function.
                    widget = self._create_widget_from_field(name, attr)
                    if not widget.tooltip:
                        widget.tooltip = _tooltips.attributes.get(name, "")

                elif isinstance(attr, FunctionGui):
                    widget = attr.copy()
                    widget[0].bind(self)  # bind self to the first argument

                else:
                    # convert class method into instance method
                    widget = getattr(self, name, None)

                if isinstance(widget, BaseGui):
                    widget.__magicclass_parent__ = self
                    self.__magicclass_children__.append(widget)
                    widget._my_symbol = Symbol(name)

                if isinstance(widget, MenuGui):
                    # Add menubar to container
                    if self._menubar is None:
                        # if widget has no menubar, a new one should be created.
                        self._menubar = QMenuBar(parent=self.native)
                        if issubclass(self.__class__, MainWindow):
                            self.native: QMainWindow
                            self.native.setMenuBar(self._menubar)
                        else:
                            if hasattr(self._widget, "_scroll_area"):
                                _layout: QBoxLayout = self._widget._qwidget.layout()
                            else:
                                _layout = self._widget._layout
                            if _layout.menuBar() is None:
                                _layout.setMenuBar(self._menubar)
                            else:
                                raise RuntimeError(
                                    "Cannot add menubar after adding a toolbar in a non-main window. "
                                    "Use maigcclass(widget_type='mainwindow') instead, or define the "
                                    "menu class before toolbar class."
                                )

                    widget.native.setParent(self._menubar, widget.native.windowFlags())
                    self._menubar.addMenu(widget.native).setText(
                        widget.name.replace("_", " ")
                    )
                    _hist.append((name, type(attr), "MenuGui"))

                elif isinstance(widget, ContextMenuGui):
                    # Add context menu to container
                    set_context_menu(widget, self)
                    _hist.append((name, type(attr), "ContextMenuGui"))

                elif isinstance(widget, ToolBarGui):
                    if self._toolbar is None:
                        self._toolbar = QtTabToolBar(widget.name, self.native)
                        if issubclass(self.__class__, MainWindow):
                            self.native: QMainWindow
                            self.native.addToolBar(self._toolbar)
                        else:
                            # self is not a main window object
                            # TODO: these codes are too dirty...
                            if isinstance(self, _USE_OUTER_LAYOUT):
                                _layout: QBoxLayout = self._widget._qwidget.layout()
                            else:
                                _layout = self._widget._layout
                            if _layout.menuBar() is None:
                                _layout.setMenuBar(self._toolbar)
                            else:
                                _layout.insertWidget(
                                    0, self._toolbar, alignment=Qt.AlignTop
                                )
                                self._toolbar.setContentsMargins(0, 0, 0, 0)
                                n_insert += 1

                    widget.native.setParent(self._toolbar, widget.native.windowFlags())
                    self._toolbar.addToolBar(
                        widget.native, widget.name.replace("_", " ")
                    )
                    _hist.append((name, type(attr), "ToolBarGui"))

                elif isinstance(widget, (Widget, Callable)):
                    if (not isinstance(widget, Widget)) and callable(widget):
                        # Methods or any callable objects, but FunctionGui is not included.
                        # NOTE: Here any custom callable objects could be given. Some callable
                        # objects can be incompatible (like "Signal" object in magicgui) but
                        # useful. Those callable objects should be passed from widget construction.
                        if name.startswith("_") or not get_additional_option(
                            attr, "gui", True
                        ):
                            keybinding = get_additional_option(attr, "keybinding", None)
                            if keybinding:
                                register_shortcut(
                                    keys=keybinding, parent=self.native, target=widget
                                )
                            continue
                        try:
                            widget = self._create_widget_from_method(widget)
                        except AttributeError as e:
                            warnings.warn(
                                f"Could not convert {widget!r} into a widget "
                                f"due to AttributeError: {e}",
                                UserWarning,
                            )
                            continue

                    elif hasattr(widget, _MCLS_PAREMT) or hasattr(
                        widget.__class__, _MCLS_PAREMT
                    ):
                        # magic-class has to know its parent.
                        # if __magicclass_parent__ is defined as a property, hasattr must be called
                        # with a type object (not instance).
                        widget.__magicclass_parent__ = self

                    else:
                        if not widget.name:
                            widget.name = name
                        if hasattr(widget, "text") and not widget.text:
                            widget.text = widget.name.replace("_", " ")

                    # Now, "widget" is a Widget object. Add widget in a way similar to "insert" method
                    # of Container.
                    if name.startswith("_"):
                        continue

                    moveto = get_additional_option(attr, "into")
                    copyto = get_additional_option(attr, "copyto", [])
                    if moveto is not None or copyto:
                        self._unwrap_method(name, widget, moveto, copyto)
                    else:
                        self._fast_insert(n_insert, widget)
                        n_insert += 1

                    _hist.append((name, str(type(attr)), type(widget).__name__))

            except Exception as e:
                format_error(e, _hist, name, attr)

        self._unify_label_widths()
        return None

    def _fast_insert(self, key: int, obj: Widget | Callable) -> None:
        if isinstance(obj, Callable):
            # Sometimes uses want to dynamically add new functions to GUI.
            if isinstance(obj, FunctionGui):
                if obj.parent is None:
                    f = nested_function_gui_callback(self, obj)
                    obj.called.connect(f)
                widget = obj
            else:
                from ._base import _inject_recorder

                obj = _inject_recorder(obj, is_method=False).__get__(self)
                widget = self._create_widget_from_method(obj)

            method_name = getattr(obj, "__name__", None)
            if method_name and not hasattr(self, method_name):
                object.__setattr__(self, method_name, obj)
        else:
            widget = obj

        # _hide_labels should not contain Container because some ValueWidget like widgets
        # are Containers.
        _hide_labels = (
            _LabeledWidget,
            ButtonWidget,
            ClassGuiBase,
            FreeWidget,
            Label,
            Image,
            Table,
            FunctionGui,
        )

        if isinstance(widget, (ValueWidget, ContainerWidget)):
            widget.changed.connect(lambda: self.changed.emit(self))

        if hasattr(widget, _MCLS_PAREMT) or hasattr(widget.__class__, _MCLS_PAREMT):
            widget.__magicclass_parent__ = self
            if isinstance(widget, ClassGuiBase):
                if self._remove_child_margins:
                    widget.margins = (0, 0, 0, 0)
                if (
                    len(self.__magicclass_children__) > 0
                    and widget is not self.__magicclass_children__[-1]
                ):
                    # nested magic classes are already in the list
                    self.__magicclass_children__.append(widget)
                    widget._my_symbol = Symbol(widget.name)

                # NOTE: This is not safe. Attributes could collision and macro recording
                # may break if not correctly named.

        _widget = widget

        if self.labels:
            # no labels for button widgets (push buttons, checkboxes, have their own)
            if not isinstance(widget, _hide_labels):
                _widget = _LabeledWidget(widget)
                widget.label_changed.connect(self._unify_label_widths)

        if key < 0:
            key += len(self)
        self._list.insert(key, widget)
        self._widget._mgui_insert_widget(key, _widget)
        return None


_C = TypeVar("_C", bound=ContainerWidget)


def make_gui(container: type[_C], no_margin: bool = True) -> type[_C | ClassGuiBase]:
    """
    Make a ClassGui class from a Container widget.
    Because GUI class inherits Container here, functions that need overriden must be defined
    here, not in ClassGuiBase.
    """

    def wrapper(cls_: type[ClassGuiBase]):
        cls: type[_C | ClassGuiBase] = type(
            cls_.__name__, (container, ClassGuiBase), {}
        )

        def __init__(
            self: cls,
            layout: str = "vertical",
            close_on_run: bool = None,
            popup_mode: str | PopUpMode = None,
            error_mode: str | ErrorMode = None,
            labels: bool = True,
            name: str = None,
            visible: bool = None,
        ):

            container.__init__(
                self, layout=layout, labels=labels, name=name, visible=visible
            )
            BaseGui.__init__(
                self,
                close_on_run=close_on_run,
                popup_mode=popup_mode,
                error_mode=error_mode,
            )

            self._menubar = None
            self._toolbar = None

            self.native.setObjectName(self.name)
            self.native.setWindowTitle(self.name)

        # ui["x"] will not return widget if x is a MagicValueField.
        # To ensure __getitem__ returns a Widget, this method should be overriden.
        def __getitem__(self: cls, key):
            """Get item by integer, str, or slice."""
            if isinstance(key, str):
                for widget in self._list:
                    if key == widget.name:
                        return widget
            return container.__getattr__(self, key)

        def __setattr__(self: cls, name: str, value: Any) -> None:
            if not isinstance(getattr(self.__class__, name, None), MagicField):
                container.__setattr__(self, name, value)
            else:
                object.__setattr__(self, name, value)

        def insert(self: cls, key: int, widget: Widget) -> None:
            self._fast_insert(key, widget)
            self._unify_label_widths()
            return None

        def reset_choices(self: cls, *_: Any):
            """Reset child Categorical widgets"""
            all_widgets: set[Widget] = set()

            for item in self._list:
                widget = getattr(item, "_inner_widget", item)
                all_widgets.add(widget)
            for widget in self.__magicclass_children__:
                all_widgets.add(widget)

            for w in all_widgets:
                if hasattr(w, "reset_choices"):
                    w.reset_choices()
            return None

        def show(self: cls, run: bool = True) -> None:
            """
            Show GUI. If any of the parent GUI is a dock widget in napari, then this
            will also show up as a dock widget (floating if in popup mode).

            Parameters
            ----------
            run : bool, default is True
                *Unlike magicgui, this parameter should always be True* unless you want
                to close the window immediately. If true, application gets executed if
                needed.
            """
            if self.__magicclass_parent__ is not None and self.parent is None:
                # If child magic class is closed before, we have to set parent again.
                self.native.setParent(
                    self.__magicclass_parent__.native, self.native.windowFlags()
                )

            viewer = self.parent_viewer
            if viewer is not None and self.parent is not None:
                name = self.parent.objectName()
                if name in viewer.window._dock_widgets and isinstance(
                    self.parent, QDockWidget
                ):
                    viewer.window._dock_widgets[name].show()
                else:
                    _floating = self._popup_mode == PopUpMode.popup
                    _area = "left" if _floating else "right"
                    dock = viewer.window.add_dock_widget(
                        self,
                        name=self.name.replace("_", " ").strip(),
                        area=_area,
                        allowed_areas=["left", "right"],
                    )
                    dock.setFloating(_floating)
            else:
                container.show(self, run=False)
                self.native.activateWindow()
                if run:
                    run_app()
            return None

        def close(self: cls):
            """Close GUI. if this widget is a dock widget, then also close it."""

            current_self = self._search_parent_magicclass()

            viewer = current_self.parent_viewer
            if viewer is not None:
                try:
                    viewer.window.remove_dock_widget(self.parent)
                except Exception:
                    pass

            container.close(self)

            return None

        if issubclass(container, MainWindow):
            # Similar to napari's viewer.window.add_dock_widget.
            # See napari/_qt/widgets/qt_viewer_dock_widget.py
            from ..wrappers import nogui

            # This function will be detected as non-reserved method so that magicclass will
            # try to convert it into widget. Should decorate with @nogui.
            @nogui
            def add_dock_widget(
                self: cls,
                widget: Widget,
                *,
                name: str = "",
                area: str = "right",
                allowed_areas: Sequence[str] | None = None,
            ):
                """
                Add a widget as a dock widget of the main window.
                This method follows napari's "add_dock_widget" method.

                Parameters
                ----------
                widget : Widget
                    Widget that will be converted into a dock widget and added to the main
                    window.
                name : str, optional
                    Name of the dock widget.
                area : str, default is "right"
                    Initial dock widget area.
                allowed_areas : sequence of str, optional
                    Allowed dock widget area. Allow all areas by default.
                """
                from ._dock_widget import QtDockWidget

                name = name or widget.name
                dock = QtDockWidget(
                    self.native,
                    widget.native,
                    name=name.replace("_", " "),
                    area=area,
                    allowed_areas=allowed_areas,
                )

                self.native.addDockWidget(QtDockWidget.areas[area], dock)
                if isinstance(widget, BaseGui):
                    widget.__magicclass_parent__ = self
                    self.__magicclass_children__.append(widget)
                    widget._my_symbol = Symbol(name)

            @nogui
            def remove_dock_widget(self: cls, widget: Widget):
                from ._dock_widget import QtDockWidget

                dock = None
                i_dock = -1
                for i, child in enumerate(self.__magicclass_children__):
                    if child is widget:
                        dock = child.native.parent()
                        if not isinstance(dock, QtDockWidget):
                            dock = None
                        i_dock = i
                        self.__magicclass_parent__ = None
                        break
                else:
                    raise RuntimeError("Dock widget not found.")

                self.native.removeDockWidget(dock)
                self.__magicclass_children__.pop(i_dock)

            @property
            def status(self: cls) -> str:
                """Get status tip."""
                return self.native.statusTip()

            @status.setter
            def status(self: cls, text: str):
                """Set status tip."""
                self.native.setStatusTip(text)
                self.native.statusBar().showMessage(text, 5000)

            cls.add_dock_widget = add_dock_widget
            cls.remove_dock_widget = remove_dock_widget
            cls.status = status

        cls.__init__ = __init__
        cls.__setattr__ = __setattr__
        cls.insert = insert
        cls.show = show
        cls.reset_choices = reset_choices
        cls.close = close
        cls._container_widget = container
        cls._remove_child_margins = no_margin
        return cls

    return wrapper


@make_gui(Container)
class ClassGui:
    pass


@make_gui(SplitterContainer)
class SplitClassGui:
    pass


@make_gui(ScrollableContainer)
class ScrollableClassGui:
    pass


@make_gui(DraggableContainer)
class DraggableClassGui:
    pass


@make_gui(CollapsibleContainer)
class CollapsibleClassGui:
    pass


@make_gui(HCollapsibleContainer)
class HCollapsibleClassGui:
    pass


@make_gui(ButtonContainer)
class ButtonClassGui:
    pass


@make_gui(ToolBoxContainer, no_margin=False)
class ToolBoxClassGui:
    pass


@make_gui(TabbedContainer, no_margin=False)
class TabbedClassGui:
    pass


@make_gui(StackedContainer, no_margin=False)
class StackedClassGui:
    pass


@make_gui(ListContainer, no_margin=False)
class ListClassGui:
    pass


@make_gui(SubWindowsContainer, no_margin=False)
class SubWindowsClassGui:
    pass


@make_gui(GroupBoxContainer, no_margin=False)
class GroupBoxClassGui:
    pass


@make_gui(FrameContainer, no_margin=False)
class FrameClassGui:
    pass


@make_gui(MainWindow)
class MainWindowClassGui:
    pass



from __future__ import annotations
from enum import Enum
import typing
from typing import Any, Union, Iterable, overload, TypeVar, Callable
from typing_extensions import Literal, Annotated, ParamSpec, _AnnotatedAlias
from magicgui.signature import split_annotated_type
from magicgui.widgets import Widget, EmptyWidget

from .fields import MagicField

try:
    from typing import _tp_cache
except ImportError:
    _tp_cache = lambda x: x

__all__ = ["WidgetType", "bound", "Bound", "Choices", "OneOf", "SomeOf", "Optional"]


class WidgetType(Enum):
    none = "none"
    scrollable = "scrollable"
    draggable = "draggable"
    split = "split"
    collapsible = "collapsible"
    hcollapsible = "hcollapsible"
    button = "button"
    toolbox = "toolbox"
    tabbed = "tabbed"
    stacked = "stacked"
    list = "list"
    subwindows = "subwindows"
    groupbox = "groupbox"
    frame = "frame"
    mainwindow = "mainwindow"


WidgetTypeStr = Union[
    Literal["none"],
    Literal["scrollable"],
    Literal["draggable"],
    Literal["split"],
    Literal["collapsible"],
    Literal["button"],
    Literal["toolbox"],
    Literal["tabbed"],
    Literal["stacked"],
    Literal["list"],
    Literal["subwindows"],
    Literal["groupbox"],
    Literal["frame"],
    Literal["mainwindow"],
    Literal["hcollapsible"],
]


PopUpModeStr = Union[
    Literal["popup"],
    Literal["first"],
    Literal["last"],
    Literal["above"],
    Literal["below"],
    Literal["dock"],
    Literal["dialog"],
    Literal["parentlast"],
]


ErrorModeStr = Union[
    Literal["msgbox"],
    Literal["stderr"],
    Literal["stdout"],
]

Color = Union[Iterable[float], str]

_W = TypeVar("_W", bound=Widget)
_V = TypeVar("_V", bound=object)
_P = ParamSpec("_P")


@overload
def bound(obj: Callable[[_W], _V]) -> type[_V]:
    ...


@overload
def bound(obj: Callable[[Any, _W], _V]) -> type[_V]:
    ...


@overload
def bound(obj: MagicField[_W, _V]) -> type[_V]:
    ...


@overload
def bound(obj: type[_W]) -> type:
    ...


def bound(obj):
    """Function version of Bound[...]."""
    # NOTE: This could be more useful than Bound??
    if callable(obj):
        outtype = obj.__annotations__.get("return", Any)
    elif isinstance(obj, MagicField):
        outtype = obj.annotation or Any
    elif isinstance(obj, (type, str)):
        outtype = Any
    else:
        raise TypeError("'bound' can only convert callable, MagicField or type objects")
    while isinstance(outtype, _AnnotatedAlias):
        outtype, _ = split_annotated_type(outtype)
    return Annotated[outtype, {"bind": obj, "widget_type": EmptyWidget}]


class _BoundAlias(type):
    """
    This metaclass is necessary for ``mypy`` to reveal type.

    For instance, if type annotation is added like this

    .. code-block:: python

        def _get_int(self, _=None) -> int:
            return 0

        def func(self, x: Bound[_get_int]):
            # do something

    ``x`` will be considered to be ``Bound`` type otherwise.
    """

    @overload
    def __getitem__(cls, value: MagicField[_W, _V]) -> type[_V]:
        ...

    @overload
    def __getitem__(cls, value: Callable[_P, _V]) -> type[_V]:
        ...

    @overload
    def __getitem__(cls, value: type[_V]) -> type[_V]:
        ...

    @_tp_cache
    def __getitem__(cls, value):
        if isinstance(value, tuple):
            raise TypeError(
                "Bound[...] should be used with only one "
                "argument (the object to be bound)."
            )
        return bound(value)


class Bound(metaclass=_BoundAlias):
    """
    Make Annotated type from a MagicField or a method, such as:

    .. code-block:: python

        from magicclass import magicclass, field

        @magicclass
        class MyClass:
            i = field(int)
            def func(self, v: Bound[i]):
                ...

    ``Bound[value]`` is identical to ``Annotated[Any, {"bind": value}]``.
    """

    def __new__(cls, *args):
        raise TypeError(
            "`Bound(...)` is deprecated since 0.5.21. Bound is now a generic alias instead "
            "of a function. Please use `Bound[...]`."
        )

    def __init_subclass__(cls, *args, **kwargs):
        raise TypeError(f"Cannot subclass {cls.__module__}.Bound")


class _OneOfAlias(type):
    """metaclass of ``OneOf``."""

    @overload
    def __getitem__(cls, value: Callable[_P, Iterable[tuple[Any, _V]]]) -> type[_V]:
        ...

    @overload
    def __getitem__(cls, value: Callable[_P, Iterable[_V]]) -> type[_V]:
        ...

    @overload
    def __getitem__(cls, value: Iterable[tuple[Any, _V]]) -> type[_V]:
        ...

    @overload
    def __getitem__(cls, value: Iterable[_V]) -> type[_V]:
        ...

    @overload
    def __getitem__(cls, value: slice) -> type[int | float]:
        ...

    @_tp_cache
    def __getitem__(cls, value):
        if callable(value):
            outtype = value.__annotations__.get("return", Any)
        elif hasattr(value, "__iter__"):
            outtype = Any
        elif isinstance(value, slice):
            outtype, value = _normalize_slice(value)
        else:
            raise TypeError("'bound' can only convert callable or iterable objects.")
        return Annotated[outtype, {"choices": value, "nullable": False}]


def _normalize_slice(value: slice) -> type | list:
    start, stop, step = value.start or 0, value.stop or 0, value.step or 1
    if float in [type(start), type(stop), type(step)]:
        import math

        ndigits = -int(min(math.log10(start), math.log10(stop), math.log10(step))) + 4
        outtype = float
        outvalue: list[float] = []
        if step > 0:
            x = start
            while x < stop:
                outvalue.append(x)
                x = round(x + step, ndigits)
        else:
            x = stop
            while start < x:
                outvalue.append(x)
                x = round(x + step, ndigits)

    else:
        outtype = int
        outvalue = list(range(start, stop, step))
    return outtype, outvalue


class OneOf(metaclass=_OneOfAlias):
    """
    Make Annotated type from a method, such as:

    .. code-block:: python

        from magicclass import magicclass

        @magicclass
        class MyClass:
            def func(self, v: OneOf[(1, 2, 3)]):
                ...

    ``OneOf[value]`` is identical to ``Annotated[Any, {"choices": value}]``.
    """

    def __new__(cls, *args):
        raise TypeError(
            "`Bound(...)` is deprecated since 0.5.21. Bound is now a generic alias instead "
            "of a function. Please use `Bound[...]`."
        )

    def __init_subclass__(cls, *args, **kwargs):
        raise TypeError(f"Cannot subclass {cls.__module__}.{cls.__name__}")


Choices = OneOf  # alias


class _SomeOfAlias(type):
    """
    This metaclass is necessary for ``mypy`` to reveal type.
    """

    @overload
    def __getitem__(
        cls, value: Callable[_P, Iterable[tuple[Any, _V]]]
    ) -> type[list[_V]]:
        ...

    @overload
    def __getitem__(cls, value: Callable[_P, Iterable[_V]]) -> type[list[_V]]:
        ...

    @overload
    def __getitem__(cls, value: Iterable[tuple[Any, _V]]) -> type[list[_V]]:
        ...

    @overload
    def __getitem__(cls, value: Iterable[_V]) -> type[list[_V]]:
        ...

    @overload
    def __getitem__(cls, value: slice) -> type[int | float]:
        ...

    @_tp_cache
    def __getitem__(cls, value):
        if callable(value):
            outtype = value.__annotations__.get("return", Any)
        elif hasattr(value, "__iter__"):
            outtype = Any
        elif isinstance(value, slice):
            outtype, value = _normalize_slice(value)
        else:
            raise TypeError("'bound' can only convert callable or iterable objects.")
        return Annotated[
            outtype, {"choices": value, "nullable": False, "widget_type": "Select"}
        ]


class SomeOf(metaclass=_SomeOfAlias):
    """
    Make Annotated type from a method, such as:

    .. code-block:: python

        from magicclass import magicclass

        @magicclass
        class MyClass:
            def func(self, v: Choices[(1, 2, 3)]):
                ...

    ``Choices[value]`` is identical to ``Annotated[Any, {"choices": value}]``.
    """

    def __new__(cls, *args):
        raise TypeError(
            "`Bound(...)` is deprecated since 0.5.21. Bound is now a generic alias instead "
            "of a function. Please use `Bound[...]`."
        )

    def __init_subclass__(cls, *args, **kwargs):
        raise TypeError(f"Cannot subclass {cls.__module__}.SomeOf")


_T = TypeVar("_T")


class _OptionalAlias(type):
    @overload
    def __getitem__(cls, value: type[_T]) -> type[typing.Optional[_T]]:
        ...

    @overload
    def __getitem__(
        cls, value: tuple[type[_T], dict[str, Any]]
    ) -> type[typing.Optional[_T]]:
        ...

    def __getitem__(cls, value):
        if not isinstance(value, (type, typing._GenericAlias)):
            raise TypeError(
                "The first argument of Optional must be a type but "
                f"got {type(value)}."
            )
        from .widgets import OptionalWidget

        opt = dict(widget_type=OptionalWidget)
        if isinstance(value, _AnnotatedAlias):
            type0, opt0 = split_annotated_type(value)
            type_ = typing.Optional[type0]
            opt.update(annotation=type_, options=opt0)
            return Annotated[type_, opt]
        else:
            opt.update(annotation=typing.Optional[value])
            return Annotated[typing.Optional[value], opt]


class Optional(metaclass=_OptionalAlias):
    """
    Make Annotated type similar to ``typing.Optional``.

    Arguments annotated with ``Optional[int]`` will create a
    ``OptionalWidget`` with a ``SpinBox`` as an inner widget.

    Examples
    --------

    from magicclass import magicclass, set_options
    from magicclass.types import Optional

    @magicclass
    class A:
        @set_options(a={"options": {"min": -1}})
        def f(self, a: Optional[int]):
            print(a)

    ui = A()
    ui.show()
    """

    def __new__(cls, *args, **kwargs):
        raise TypeError("Type Optional cannot be instantiated.")

    def __init_subclass__(cls, *args, **kwargs):
        raise TypeError(f"Cannot subclass {cls.__module__}.Optional.")


def __getattr__(key: str):
    if key in ["List", "Tuple"]:
        import warnings

        warnings.warn(
            f"Type {key!r} is deprecated. Please use typing.{key}.",
            DeprecationWarning,
            stacklevel=2,
        )
        import typing

        return getattr(typing, key)
    raise AttributeError(f"module {__name__!r} has no attribute {key!r}")



from napari import Viewer
from napari.layers import Layer
from macrokit import register_type, Symbol, Expr

VIEWER_SYMBOL = Symbol.var("viewer")

# NOTE: To support macro recording when multiple viewer exists, calling
# napari.current_viewer() is not safe if a new viewer is created during a function
# call. Since standard Python script does not distinguish different viewers, here
# we simply let "viewer" to represent any viewer.
register_type(Viewer, lambda viewer: VIEWER_SYMBOL)


@register_type(Layer)
def _get_layer(layer: Layer):
    """Record layer object as ``viewer.layers["layer-name"]``"""

    expr = Expr("getitem", [Expr("getattr", [VIEWER_SYMBOL, "layers"]), layer.name])
    return expr



from __future__ import annotations
from functools import wraps as functools_wraps
from typing import (
    Any,
    Callable,
    TYPE_CHECKING,
    Iterable,
    Iterator,
    TypeVar,
    overload,
    MutableSequence,
)
from types import MethodType
from abc import ABCMeta
from typing_extensions import _AnnotatedAlias, Literal
import inspect
import warnings
import os
from enum import Enum
import warnings
from docstring_parser import parse, compose
from qtpy.QtWidgets import QWidget, QDockWidget
from qtpy.QtGui import QIcon

from psygnal import Signal
from magicgui.signature import MagicParameter, split_annotated_type
from magicgui.widgets import (
    FunctionGui,
    FileEdit,
    EmptyWidget,
    Widget,
    Container,
    Image,
    Table,
    Label,
    MainWindow,
)
from magicgui.application import use_app
from magicgui.widgets._bases.widget import Widget
from magicgui.widgets._bases import ButtonWidget, ValueWidget
from macrokit import Expr, Head, Symbol, symbol

from .keybinding import as_shortcut
from .mgui_ext import (
    AbstractAction,
    Action,
    FunctionGuiPlus,
    PushButtonPlus,
    _LabeledWidgetAction,
    mguiLike,
)
from .utils import copy_class, get_parameters, callable_to_classes
from ._macro import GuiMacro

from ..utils import (
    get_signature,
    iter_members,
    Tooltips,
    move_to_screen_center,
    argcount,
    is_instance_method,
    method_as_getter,
    eval_attribute,
    thread_worker,
)
from ..widgets import Separator, FreeWidget
from ..fields import MagicField
from ..signature import MagicMethodSignature, get_additional_option
from ..wrappers import upgrade_signature

if TYPE_CHECKING:
    import numpy as np
    import napari
    from types import TracebackType


class PopUpMode(Enum):
    """Define how to popup FunctionGui."""

    popup = "popup"
    first = "first"
    last = "last"
    above = "above"
    below = "below"
    dock = "dock"
    dialog = "dialog"
    parentlast = "parentlast"


def _msgbox_raising(e, parent):
    from ._message_box import QtErrorMessageBox

    return QtErrorMessageBox.raise_(e, parent=parent.native)


def _stderr_raising(e, parent):
    pass


def _stdout_raising(e, parent):
    print(f"{e.__class__.__name__}: {e}")


def _napari_notification_raising(e, parent):
    from napari.utils.notifications import show_error

    show_error(str(e))


class ErrorMode(Enum):
    msgbox = "msgbox"
    stderr = "stderr"
    stdout = "stdout"
    napari = "napari"

    def get_handler(self):
        """Get error handler."""
        return ErrorModeHandlers[self]

    def wrap_handler(self, func: Callable, parent):
        """Wrap function with the error handler."""
        handler = self.get_handler()

        @functools_wraps(func)
        def wrapped_func(*args, **kwargs):
            try:
                out = func(*args, **kwargs)
            except Exception as e:
                e.__traceback__ = _cleanup_tb(e.__traceback__)
                handler(e, parent=parent)
                out = e
            return out

        return wrapped_func


ErrorModeHandlers = {
    ErrorMode.msgbox: _msgbox_raising,
    ErrorMode.stderr: _stderr_raising,
    ErrorMode.stdout: _stdout_raising,
    ErrorMode.napari: _napari_notification_raising,
}


defaults = {
    "popup_mode": PopUpMode.popup,
    "error_mode": ErrorMode.msgbox,
    "close_on_run": True,
    "macro-max-history": 1000,
}

_RESERVED = {
    "__magicclass_parent__",
    "__magicclass_children__",
    "_close_on_run",
    "_error_mode",
    "_popup_mode",
    "_my_symbol",
    "_macro_instance",
    "macro",
    "annotation",
    "enabled",
    "find_ancestor",
    "gui_only",
    "height",
    "label_changed",
    "label",
    "layout",
    "labels",
    "margins",
    "max_height",
    "max_width",
    "min_height",
    "min_width",
    "name",
    "options",
    "param_kind",
    "parent_changed",
    "tooltip",
    "visible",
    "widget_type",
    "width",
    "wraps",
    "_unwrap_method",
    "_search_parent_magicclass",
    "_iter_child_magicclasses",
}


def check_override(cls: type):
    """
    Some of the methods should not be overriden because they are essential for magic
    class construction.

    Parameters
    ----------
    cls : type
        Base class to test override.

    Raises
    ------
    AttributeError
        If forbidden override found.
    """
    subclass_members = set(cls.__dict__.keys())
    collision = subclass_members & _RESERVED
    if collision:
        raise AttributeError(
            f"Cannot override magic class reserved attributes: {collision}"
        )


_T = TypeVar("_T", bound="MagicTemplate")
_F = TypeVar("_F", bound=Callable)


class _MagicTemplateMeta(ABCMeta):
    """This metaclass enables type checking of nested magicclasses."""

    @overload
    def __get__(self: type[_T], obj: Any, objtype=None) -> _T:
        ...

    @overload
    def __get__(self: type[_T], obj: Literal[None], objtype=None) -> type[_T]:
        ...

    def __get__(self, obj, objtype=None):
        return self


class MagicTemplate(metaclass=_MagicTemplateMeta):
    __doc__ = ""
    __magicclass_parent__: None | MagicTemplate
    __magicclass_children__: list[MagicTemplate]
    _close_on_run: bool
    _component_class: type[Action | Widget]
    _error_mode: ErrorMode
    _list: list[Action | Widget]
    _macro_instance: GuiMacro
    _my_symbol: Symbol
    _popup_mode: PopUpMode
    annotation: Any
    changed: Signal
    enabled: bool
    gui_only: bool
    height: int
    icon_path: str
    label_changed: Signal
    label: str
    layout: str
    labels: bool
    margins: tuple[int, int, int, int]
    max_height: int
    max_width: int
    min_height: int
    min_width: int
    name: str
    native: QWidget
    options: dict
    param_kind: inspect._ParameterKind
    parent: Widget
    parent_changed: Signal
    tooltip: str
    visible: bool
    widget_type: str
    width: int

    __init_subclass__ = check_override

    def show(self, run: bool) -> None:
        raise NotImplementedError()

    def hide(self) -> None:
        raise NotImplementedError()

    def close(self) -> None:
        raise NotImplementedError()

    @overload
    def __getitem__(self, key: int | str) -> Widget:
        ...

    @overload
    def __getitem__(self, key: slice) -> MutableSequence[Widget]:
        ...

    def __getitem__(self, key):
        raise NotImplementedError()

    def index(self, value: Any, start: int, stop: int) -> int:
        raise NotImplementedError()

    def remove(self, value: Widget | str):
        raise NotImplementedError()

    def append(self, widget: Widget) -> None:
        return self.insert(len(self, widget))

    def _fast_insert(self, key: int, widget: Widget) -> None:
        raise NotImplementedError()

    def insert(self, key: int, widget: Widget) -> None:
        self._fast_insert(key, widget)
        self._unify_label_widths()

    def render(self) -> np.ndarray:
        raise NotImplementedError()

    def _unify_label_widths(self):
        raise NotImplementedError()

    def reset_choices(self, *args):
        raise NotImplementedError()

    @property
    def macro(self) -> GuiMacro:
        """The macro object bound to the ancestor widget."""
        if self.__magicclass_parent__ is None:
            return self._macro_instance
        else:
            return self.__magicclass_parent__.macro

    @property
    def parent_viewer(self) -> napari.Viewer | None:
        """Return napari.Viewer if magic class is a dock widget of a viewer."""
        parent_self = self._search_parent_magicclass()
        if parent_self.native.parent() is None:
            return None
        try:
            from napari.utils._magicgui import find_viewer_ancestor
        except ImportError:
            return None
        viewer = find_viewer_ancestor(parent_self.native)
        return viewer

    @property
    def parent_dock_widget(self) -> QDockWidget | None:
        """
        Return dock widget object if magic class is a dock widget of a main
        window widget, such as a napari Viewer.
        """
        parent_self = self._search_parent_magicclass()
        try:
            dock = parent_self.native.parent()
            if not isinstance(dock, QDockWidget):
                dock = None
        except AttributeError:
            dock = None

        return dock

    def find_ancestor(self, ancestor: type[_T]) -> _T:
        """
        Find magic class ancestor whose type matches the input.
        This method is useful when a child widget class is defined outside a magic
        class while it needs access to its parent.

        Parameters
        ----------
        ancestor : type of MagicTemplate
            Type of ancestor to search for.

        Returns
        -------
        MagicTemplate
            Magic class object if found.
        """
        if not isinstance(ancestor, type):
            raise TypeError(
                "The first argument of 'find_ancestor' must be a type but got "
                f"{type(ancestor)}"
            )

        current_self = self
        while type(current_self) is not ancestor:
            current_self = current_self.__magicclass_parent__
            if current_self is None:
                raise RuntimeError(
                    f"Magic class {ancestor.__name__} not found. {ancestor.__name__} it "
                    f"is not an ancestor of {self.__class__.__name__}"
                )
        return current_self

    def objectName(self) -> str:
        """
        Return object name of the QWidget.

        This function makes the object name discoverable by napari's
        `viewer.window.add_dock_widget` function. At the same time, since this function
        will always be called when the widget is added as a dock widget of napari, we
        can import macro recorders for napari types in the appropriate timing.
        """
        try:
            from . import _napari_type  # load default macro recorder.
        except Exception:
            pass
        return self.native.objectName()

    @classmethod
    def wraps(
        cls,
        method: _F | None = None,
        *,
        template: Callable | None = None,
        copy: bool = False,
    ) -> _F:
        """
        Wrap a parent method in a child magic-class.

        Wrapped method will appear in the child widget but behaves as if it is in
        the parent widget. Basically, this function is used as a wrapper like below.

        .. code-block:: python

            @magicclass
            class C:
                @magicclass
                class D:
                    def func(self, ...): ... # pre-definition
                @D.wraps
                def func(self, ...): ...

        Parameters
        ----------
        method : Callable, optional
            Method of parent class.
        template : Callable, optional
            Function template for signature.
        copy: bool, default is False
            If true, wrapped method is still enabled.

        Returns
        -------
        Callable
            Same method as input, but has updated signature.
        """
        if (not copy) and get_additional_option(method, "into", None) is not None:
            # If method is already wrapped, wraps should create a copy.
            copy = True

        def wrapper(method: _F):
            # Base function to get access to the original function
            if isinstance(method, FunctionGui):
                func = method._function
            else:
                func = method

            if template is not None:
                wraps(template)(func)

            predefined = getattr(cls, func.__name__, None)
            if predefined is not None:
                # Update signature to the parent one. This step is necessary when widget design
                # is defined on the parent side. Parameters should be replaced with a simplest
                # one to avoid creating useless widgets.
                parent_sig = get_signature(func)
                _simple_param = inspect.Parameter(
                    "self", inspect.Parameter.POSITIONAL_OR_KEYWORD
                )
                if not hasattr(predefined, "__signature__"):
                    predefined.__signature__ = parent_sig.replace(
                        parameters=(_simple_param,),
                    )
                else:
                    sig: inspect.Signature = predefined.__signature__
                    predefined.__signature__ = sig.replace(
                        parameters=parent_sig.parameters.values(),
                        return_annotation=parent_sig.return_annotation,
                    )
                upgrade_signature(predefined, additional_options={"copyto": []})

            if copy:
                copyto_list = get_additional_option(func, "copyto", [])
                copyto_list.append(cls.__name__)
                upgrade_signature(func, additional_options={"copyto": copyto_list})
            else:
                upgrade_signature(func, additional_options={"into": cls.__name__})
            return method

        return wrapper if method is None else wrapper(method)

    def _unwrap_method(
        self,
        method_name: str,
        widget: FunctionGui | PushButtonPlus | Action,
        moveto: str,
        copyto: list[str],
    ):
        """
        This private method converts class methods that are wrapped by its child widget class
        into widget in child widget. Practically same widget is shared between parent and child,
        but only visible in the child side.

        Parameters
        ----------
        moveto : str
            Name of child widget class name.
        method_name : str
            Name of method.
        widget : FunctionGui
            Widget to be added.

        Raises
        ------
        RuntimeError
            If ``child_clsname`` was not found in child widget list. This error will NEVER be raised
            in the user's side.
        """
        if moveto is not None:
            matcher = copyto + [moveto]
        else:
            matcher = copyto

        _found = 0
        _n_match = len(matcher)

        for child_instance in self._iter_child_magicclasses():
            _name = child_instance.__class__.__name__
            if _name in matcher:
                # get the position of predefined child widget
                try:
                    index = _get_index(child_instance, method_name)
                    new = False
                except ValueError:
                    index = -1
                    new = True

                self._fast_insert(-1, widget)
                copy = _name in copyto

                if isinstance(widget, FunctionGui):
                    if copy:
                        widget = widget.copy()
                    if new:
                        child_instance._fast_insert(-1, widget)
                    else:
                        del child_instance[index]
                        child_instance._fast_insert(index, widget)

                else:
                    widget.visible = copy
                    if new:
                        child_widget = child_instance._create_widget_from_method(
                            lambda x: None
                        )
                        child_widget.text = widget.text
                        child_instance._fast_insert(-1, child_widget)
                    else:
                        child_widget: PushButtonPlus | AbstractAction = child_instance[
                            index
                        ]

                    child_widget.changed.disconnect()
                    child_widget.changed.connect(widget.changed)
                    child_widget.tooltip = widget.tooltip
                    child_widget._doc = widget._doc

                widget._unwrapped = True

                _found += 1
                if _found == _n_match:
                    break

        else:
            raise RuntimeError(
                f"{method_name} not found in class {self.__class__.__name__}"
            )

    def _convert_attributes_into_widgets(self):
        """
        This function is called in dynamically created __init__. Methods, fields and nested
        classes are converted to magicgui widgets.
        """
        raise NotImplementedError()

    def _create_widget_from_field(self, name: str, fld: MagicField):
        """
        This function is called when magic-class encountered a MagicField in its definition.

        Parameters
        ----------
        name : str
            Name of variable
        fld : MagicField
            A field object that describes what type of widget it should be.
        """
        raise NotImplementedError()

    def _create_widget_from_method(self, obj: MethodType):
        """Convert instance methods into GUI objects, such as push buttons or actions."""
        text = obj.__name__.replace("_", " ")
        widget = self._component_class(name=obj.__name__, text=text, gui_only=True)

        func = _create_gui_method(self, obj)

        # Prepare a button or action
        widget.tooltip = Tooltips(func).desc
        widget._doc = func.__doc__

        # Get the number of parameters except for empty widgets.
        # With these lines, "bind" method of magicgui works inside magicclass.
        fgui_classes = callable_to_classes(func)
        n_empty = len(
            [_wdg_cls for _wdg_cls in fgui_classes if _wdg_cls is EmptyWidget]
        )
        nparams = argcount(func) - n_empty

        has_preview = get_additional_option(func, "preview", None) is not None

        if nparams == 0 and not has_preview:
            # We don't want a dialog with a single widget "Run" to show up.
            def run_function():
                # NOTE: callback must be defined inside function. Magic class must be
                # "compiled" otherwise function wrappings are not ready!
                mgui = _build_mgui(widget, func, self)
                mgui.native.setParent(self.native, mgui.native.windowFlags())
                out = mgui()

                return out

        elif nparams == 1 and issubclass(fgui_classes[0], FileEdit) and not has_preview:
            # We don't want to open a magicgui dialog and again open a file dialog.
            def run_function():
                mgui = _build_mgui(widget, func, self)
                mgui.native.setParent(self.native, mgui.native.windowFlags())
                fdialog: FileEdit = mgui[0]
                if result := fdialog._show_file_dialog(
                    fdialog.mode,
                    caption=fdialog._btn_text,
                    start_path=str(fdialog.value),
                    filter=fdialog.filter,
                ):
                    fdialog.value = result
                    out = mgui(result)
                else:
                    out = None
                return out

        else:
            _prep_func = _define_popup(self, func, widget)

            def run_function():
                mgui = _build_mgui(widget, func, self)
                if mgui.call_count == 0 and len(mgui.called._slots) == 0:
                    _prep_func(mgui)
                    if self._popup_mode not in (
                        PopUpMode.popup,
                        PopUpMode.dock,
                        PopUpMode.dialog,
                    ):
                        mgui.label = ""
                        # to avoid name collision
                        mgui.name = f"mgui-{id(mgui._function)}"
                        mgui.margins = (0, 0, 0, 0)
                        title = Separator(
                            orientation="horizontal", title=text, button=True
                        )
                        title.btn_text = "-"
                        # TODO: should remove mgui from self?
                        title.btn_clicked.connect(mgui.hide)
                        mgui.insert(0, title)

                    if self._close_on_run and not mgui._auto_call:
                        if self._popup_mode not in (PopUpMode.dock, PopUpMode.dialog):
                            mgui.called.connect(mgui.hide)
                        elif self._popup_mode == PopUpMode.dock:
                            # If FunctioGui is docked, we should close QDockWidget.
                            mgui.called.connect(lambda: mgui.parent.hide())

                if nparams == 1 and issubclass(fgui_classes[0], FileEdit):
                    fdialog: FileEdit = mgui[0]
                    if result := fdialog._show_file_dialog(
                        fdialog.mode,
                        caption=fdialog._btn_text,
                        start_path=str(fdialog.value),
                        filter=fdialog.filter,
                    ):
                        fdialog.value = result
                    else:
                        return None

                if self._popup_mode not in (PopUpMode.dock, PopUpMode.dialog):
                    widget.mgui.show()
                elif self._popup_mode == PopUpMode.dock:
                    mgui.parent.show()  # show dock widget
                else:
                    mgui.exec_as_dialog(parent=self)

                return None

        widget.changed.connect(run_function)

        # If design is given, load the options.
        widget.from_options(func)

        # keybinding
        keybinding = get_additional_option(func, "keybinding", None)
        if keybinding is not None:
            shortcut = as_shortcut(keybinding)
            widget.set_shortcut(shortcut)

        return widget

    def _search_parent_magicclass(self) -> MagicTemplate:
        """Find the ancestor."""
        current_self = self
        while (
            parent := getattr(current_self, "__magicclass_parent__", None)
        ) is not None:
            current_self = parent
        return current_self

    def _iter_child_magicclasses(self) -> Iterable[MagicTemplate]:
        """Iterate over all the child magic classes"""
        for child in self.__magicclass_children__:
            yield child
            yield from child._iter_child_magicclasses()

    def _call_with_return_callback(self, fname: str, *args, **kwargs) -> None:
        from ..core import get_function_gui

        fgui = get_function_gui(self, fname)
        fgui(*args, **kwargs)
        return None


class BaseGui(MagicTemplate):
    def __init__(self, close_on_run, popup_mode, error_mode):
        self._macro_instance = GuiMacro(
            max_lines=defaults["macro-max-history"],
            flags={"Get": False, "Return": False},
        )
        self.__magicclass_parent__: BaseGui | None = None
        self.__magicclass_children__: list[MagicTemplate] = []
        self._close_on_run = close_on_run
        self._popup_mode = popup_mode
        self._error_mode = error_mode
        self._my_symbol = Symbol.var("ui")
        self._icon_path = None


class ContainerLikeGui(BaseGui, mguiLike, MutableSequence):
    # This class enables similar API between magicgui widgets and additional widgets
    # in magicclass such as menu and toolbar.
    _component_class = Action
    changed = Signal(object)
    _list: list[AbstractAction | ContainerLikeGui]

    @property
    def icon_path(self):
        return self._icon_path

    @icon_path.setter
    def icon_path(self, path: str):
        path = str(path)
        if os.path.exists(path):
            icon = QIcon(path)
            if hasattr(self.native, "setIcon"):
                self.native.setIcon(icon)
            else:
                self.native.setWindowIcon(icon)
            self._icon_path = path
        else:
            warnings.warn(
                f"Path {path} does not exists. Could not set icon.", UserWarning
            )

    def reset_choices(self, *_: Any):
        """Reset child Categorical widgets"""
        all_widgets: set[Widget] = set()

        for item in self._list:
            widget = getattr(item, "_inner_widget", item)
            all_widgets.add(widget)
        for widget in self.__magicclass_children__:
            all_widgets.add(widget)

        for w in all_widgets:
            if hasattr(w, "reset_choices"):
                w.reset_choices()

    def _create_widget_from_field(self, name: str, fld: MagicField):
        if fld.not_ready():
            raise TypeError(
                f"MagicField {name} does not contain enough information for widget creation"
            )

        fld.name = fld.name or name
        action = fld.get_action(self)

        if action.support_value and fld.record:
            # By default, set value function will be connected to the widget.
            getvalue = type(fld) is MagicField
            f = value_widget_callback(self, action, name, getvalue=getvalue)
            action.changed.connect(f)

        return action

    def __getitem__(self, key: int | str) -> ContainerLikeGui | AbstractAction:
        if isinstance(key, int):
            return self._list[key]

        out = None
        for obj in self._list:
            if obj.name == key:
                out = obj
                break
        if out is None:
            raise KeyError(key)
        return out

    def __setitem__(self, key, value):
        raise NotImplementedError()

    def __delitem__(self, key: int | str) -> None:
        self.native.removeAction(self[key].native)

    def __iter__(self) -> Iterator[ContainerLikeGui | AbstractAction]:
        return iter(self._list)

    def __len__(self) -> int:
        return len(self._list)

    def append(self, obj: Callable | ContainerLikeGui | AbstractAction) -> None:
        return self.insert(len(self._list), obj)

    def _unify_label_widths(self):
        _hide_labels = (
            _LabeledWidgetAction,
            ButtonWidget,
            FreeWidget,
            Label,
            FunctionGui,
            BaseGui,
            Image,
            Table,
            Action,
        )
        need_labels = [w for w in self if not isinstance(w, _hide_labels)]

        if self.labels and need_labels:
            measure = use_app().get_obj("get_text_width")
            widest_label = max(measure(w.label) for w in need_labels)
            for w in need_labels:
                labeled_widget = w._labeled_widget()
                if labeled_widget:
                    labeled_widget.label_width = widest_label

    def render(self):
        try:
            import numpy as np
        except ImportError:
            raise ModuleNotFoundError(
                "could not find module 'numpy'. "
                "Please `pip install numpy` to render widgets."
            ) from None
        import qtpy

        img = self.native.grab().toImage()
        bits = img.constBits()
        h, w, c = img.height(), img.width(), 4
        if qtpy.API_NAME == "PySide2":
            arr = np.array(bits).reshape(h, w, c)
        else:
            bits.setsize(h * w * c)
            arr = np.frombuffer(bits, np.uint8).reshape(h, w, c)

        return arr[:, :, [2, 1, 0, 3]]

    def _repr_png_(self):
        """Return PNG representation of the widget for QtConsole."""
        from io import BytesIO

        try:
            from imageio import imsave
        except ImportError:
            print(
                "(For a nicer magicmenu widget representation in "
                "Jupyter, please `pip install imageio`)"
            )
            return None

        with BytesIO() as file_obj:
            imsave(file_obj, self.render(), format="png")
            file_obj.seek(0)
            return file_obj.read()


def _get_widget_name(widget: Widget):
    # To escape reference
    return widget.name


def _create_gui_method(self: BaseGui, obj: MethodType):
    func_sig = inspect.signature(obj)
    # Method type cannot set __signature__ attribute.
    @functools_wraps(obj)
    def func(*args, **kwargs):
        return obj(*args, **kwargs)

    func.__signature__ = func_sig

    # This block enables instance methods in "bind" or "choices" of ValueWidget.
    all_params: list[inspect.Parameter] = []
    for param in func.__signature__.parameters.values():
        if isinstance(param.annotation, _AnnotatedAlias):
            param = MagicParameter.from_parameter(param)

        if isinstance(param, MagicParameter):
            _param = MagicParameter(
                name=param.name,
                default=param.default,
                annotation=split_annotated_type(param.annotation)[0],
                gui_options=param.options.copy(),
            )
            _arg_bind = _param.options.get("bind", None)
            _arg_choices = _param.options.get("choices", None)

            # If bound method is a class method, use self.method(widget).
            if isinstance(_arg_bind, str):
                _arg_bind = eval_attribute(type(self), _arg_bind)

            if is_instance_method(_arg_bind):
                _param.options["bind"] = method_as_getter(self, _arg_bind)

            # If a MagicFiled is bound, bind the value of the connected widget.
            elif isinstance(_arg_bind, MagicField):
                _param.options["bind"] = _arg_bind.as_remote_getter(self)

            # If choices are provided by a class method, use self.method(widget).
            if isinstance(_arg_choices, str):
                _arg_choices = eval_attribute(type(self), _arg_choices)

            if is_instance_method(_arg_choices):
                _param.options["choices"] = method_as_getter(self, _arg_choices)

        else:
            _param = param

        all_params.append(_param)

    func.__signature__ = func.__signature__.replace(parameters=all_params)
    if isinstance(func.__signature__, MagicMethodSignature):
        func.__signature__.additional_options = getattr(
            func_sig, "additional_options", {}
        )
    return func


def _build_mgui(widget_: Action | PushButtonPlus, func: Callable, parent: BaseGui):
    if widget_.mgui is not None:
        return widget_.mgui
    try:
        sig = getattr(func, "__signature__", None)
        if isinstance(sig, MagicMethodSignature):
            opt = sig.additional_options
        else:
            opt = {}

        # confirmation
        conf = opt.get("confirm", None)
        if conf is not None:
            func = _implement_confirmation(func, parent, **conf)

        # Wrap function to deal with errors in a right way.
        func = parent._error_mode.wrap_handler(func, parent=parent)

        call_button = opt.get("call_button", None)
        layout = opt.get("layout", "vertical")
        labels = opt.get("labels", True)
        auto_call = opt.get("auto_call", False)
        mgui = FunctionGuiPlus(
            func, call_button, layout=layout, labels=labels, auto_call=auto_call
        )
        preview = opt.get("preview", None)
        if preview is not None:
            btn_text, previewer = preview
            mgui.append_preview(previewer.__get__(parent), btn_text)

    except Exception as e:
        msg = (
            "Exception was raised during building magicgui from method "
            f"{func.__name__}.\n{e.__class__.__name__}: {e}"
        )
        raise type(e)(msg)

    widget_.mgui = mgui
    name = widget_.name or ""
    mgui.native.setWindowTitle(name.replace("_", " ").strip())
    return mgui


_C = TypeVar("_C", Callable, type)


def wraps(template: Callable | inspect.Signature) -> Callable[[_C], _C]:
    """
    Update signature using a template. If class is wrapped, then all the methods
    except for those start with "__" will be wrapped.

    Parameters
    ----------
    template : Callable or inspect.Signature object
        Template function or its signature.

    Returns
    -------
    Callable
        A wrapper which take a function or class as an input and returns same
        function or class with updated signature(s).
    """

    def wrapper(f: _C) -> _C:
        if isinstance(f, type):
            for name, attr in iter_members(f):
                if callable(attr) or isinstance(attr, type):
                    wrapper(attr)
            return f

        Param = inspect.Parameter
        old_signature = inspect.signature(f)

        old_params = old_signature.parameters

        if callable(template):
            template_signature = inspect.signature(template)
        elif isinstance(template, inspect.Signature):
            template_signature = template
        else:
            raise TypeError(
                "template must be a callable object or signature, "
                f"but got {type(template)}."
            )

        # update empty signatures
        template_params = template_signature.parameters
        new_params: list[Param] = []

        for k, v in old_params.items():
            if v.annotation is Param.empty and v.default is Param.empty:
                new_params.append(
                    template_params.get(k, Param(k, Param.POSITIONAL_OR_KEYWORD))
                )
            else:
                new_params.append(v)

        # update empty return annotation
        if old_signature.return_annotation is inspect.Parameter.empty:
            return_annotation = template_signature.return_annotation
        else:
            return_annotation = old_signature.return_annotation

        f.__signature__ = inspect.Signature(
            parameters=new_params, return_annotation=return_annotation
        )

        fdoc = parse(f.__doc__)
        tempdoc = parse(template.__doc__)
        fdoc.short_description = fdoc.short_description or tempdoc.short_description
        fdoc.long_description = fdoc.long_description or tempdoc.long_description
        fdoc.meta = fdoc.meta or tempdoc.meta
        f.__doc__ = compose(fdoc)

        return f

    return wrapper


def _get_index(container: Container, widget_or_name: Widget | str) -> int:
    """
    Identical to container[widget_or_name], which sometimes doesn't work
    in magic-class.
    """
    if isinstance(widget_or_name, str):
        name = widget_or_name
    else:
        name = widget_or_name.name
    for i, widget in enumerate(container):
        if widget.name == name:
            break
    else:
        raise ValueError(f"{widget_or_name} not found in {container}")
    return i


def _child_that_has_widget(
    self: BaseGui, method: Callable, widget_or_name: Widget | str
) -> BaseGui:
    child_clsname = get_additional_option(method, "into")
    if child_clsname is None:
        return self
    for child_instance in self._iter_child_magicclasses():
        if child_instance.__class__.__name__ == child_clsname:
            break
    else:
        raise ValueError(f"{widget_or_name} not found.")
    return child_instance


def value_widget_callback(
    gui: MagicTemplate,
    widget: ValueWidget,
    name: str | list[str],
    getvalue: bool = True,
):
    """Define a ValueWidget callback, including macro recording."""
    if isinstance(name, str):
        sym_name = Symbol(name)
    else:
        sym_name = Symbol(name[0])
        for n in name[1:]:
            sym_name = Expr(head=Head.getattr, args=[sym_name, n])

    if getvalue:
        sub = Expr(head=Head.getattr, args=[sym_name, Symbol("value")])  # name.value
    else:
        sub = sym_name

    def _set_value():
        if not widget.enabled or not gui.macro.active:
            # If widget is read only, it means that value is set in script (not manually).
            # Thus this event should not be recorded as a macro.
            return None

        gui.changed.emit(gui)

        # Make an expression of
        # >>> x.name.value = value
        # or
        # >>> x.name = value
        target = Expr(Head.getattr, [symbol(gui), sub])
        expr = Expr(Head.assign, [target, widget.value])
        if gui.macro._last_setval == target and len(gui.macro) > 0:
            gui.macro.pop()
            gui.macro._erase_last()
        else:
            gui.macro._last_setval = target
        gui.macro.append(expr)
        return None

    return _set_value


def nested_function_gui_callback(gui: MagicTemplate, fgui: FunctionGui):
    """Define a FunctionGui callback, including macro recording."""
    fgui_name = Symbol(fgui.name)

    def _after_run():
        if not fgui.enabled or not gui.macro.active:
            # If widget is read only, it means that value is set in script (not manually).
            # Thus this event should not be recorded as a macro.
            return None
        inputs = get_parameters(fgui)
        args = [Expr(head=Head.kw, args=[Symbol(k), v]) for k, v in inputs.items()]
        # args[0] is self
        sub = Expr(head=Head.getattr, args=[symbol(gui), fgui_name])  # {x}.func
        expr = Expr(head=Head.call, args=[sub] + args[1:])  # {x}.func(args...)

        if fgui._auto_call:
            # Auto-call will cause many redundant macros. To avoid this, only the last input
            # will be recorded in magic-class.
            last_expr = gui.macro[-1]
            if (
                last_expr.head == Head.call
                and last_expr.args[0].head == Head.getattr
                and last_expr.at(0, 1) == expr.at(0, 1)
                and len(gui.macro) > 0
            ):
                gui.macro.pop()
                gui.macro._erase_last()

        gui.macro.append(expr)
        gui.macro._last_setval = None

    return _after_run


_SELF = inspect.Parameter("self", inspect.Parameter.POSITIONAL_OR_KEYWORD)


def _inject_recorder(func: Callable, is_method: bool = True) -> Callable:
    """Inject macro recording functionality into a function."""
    sig = get_signature(func)
    if is_method:
        sig = sig.replace(
            parameters=list(sig.parameters.values())[1:],
            return_annotation=sig.return_annotation,
        )
        _func = func
    else:

        @functools_wraps(func)
        def _func(self, *args, **kwargs):
            return func(*args, **kwargs)

        _func.__signature__ = sig.replace(
            parameters=[_SELF] + list(sig.parameters.values()),
            return_annotation=sig.return_annotation,
        )

    _record_macro = _define_macro_recorder(sig, _func)

    if not isinstance(_func, thread_worker):

        @functools_wraps(_func)
        def _recordable(bgui: MagicTemplate, *args, **kwargs):
            with bgui.macro.blocked():
                out = _func.__get__(bgui)(*args, **kwargs)
            if bgui.macro.active:
                _record_macro(bgui, *args, **kwargs)
            return out

        if hasattr(_func, "__signature__"):
            _recordable.__signature__ = _func.__signature__
        return _recordable

    else:
        _func._set_recorder(_record_macro)
        return _func


def _define_macro_recorder(sig: inspect.Signature, func: Callable):
    if isinstance(sig, MagicMethodSignature):
        opt = sig.additional_options
        _auto_call = opt.get("auto_call", False)
    else:
        _auto_call = False

    if sig.return_annotation is inspect.Parameter.empty:

        def _record_macro(bgui: MagicTemplate, *args, **kwargs):
            bound = sig.bind(*args, **kwargs)
            kwargs = dict(bound.arguments.items())
            expr = Expr.parse_method(bgui, func, (), kwargs)
            if _auto_call:
                # Auto-call will cause many redundant macros. To avoid this, only the last
                # input will be recorded in magic-class.
                last_expr = bgui.macro[-1]
                if (
                    last_expr.head == Head.call
                    and last_expr.args[0].head == Head.getattr
                    and last_expr.at(0, 1) == expr.at(0, 1)
                    and len(bgui.macro) > 0
                ):
                    bgui.macro.pop()
                    bgui.macro._erase_last()

            bgui.macro.append(expr)
            bgui.macro._last_setval = None
            return None

    else:
        _cname_ = "_call_with_return_callback"

        def _record_macro(bgui: MagicTemplate, *args, **kwargs):
            bound = sig.bind(*args, **kwargs)
            kwargs = dict(bound.arguments.items())
            expr = Expr.parse_method(bgui, _cname_, (func.__name__,), kwargs)
            if _auto_call:
                # Auto-call will cause many redundant macros. To avoid this, only the last
                # input will be recorded in magic-class.
                last_expr = bgui.macro[-1]
                if (
                    last_expr.head == Head.call
                    and last_expr.args[0].head == Head.getattr
                    and last_expr.at(0, 1) == expr.at(0, 1)
                    and last_expr.args[1] == expr.args[1]
                    and len(bgui.macro) > 0
                ):
                    bgui.macro.pop()
                    bgui.macro._erase_last()

            bgui.macro.append(expr)
            bgui.macro._last_setval = None
            return None

    return _record_macro


def convert_attributes(cls: type[_T], hide: tuple[type, ...]) -> dict[str, Any]:
    """
    Convert class attributes into macro recordable ones.

    Returned dictionary can be directly used for the third argument of
    ``type`` constructor. To avoid converting all the callables in
    subclasses, subclasses that will be iterated over can be restricted
    using ``hide`` argument.

    Parameters
    ----------
    cls : BaseGui type
        Class that will be converted.
    hide : tuple of types
        MROs that will be ignored during iteration.

    Returns
    -------
    dict
        New namespace.
    """
    _dict: dict[str, Callable] = {}
    _pass_type = (property, classmethod, staticmethod, type, Widget)
    mro = [c for c in cls.__mro__ if c not in hide]
    for subcls in reversed(mro):
        for name, obj in subcls.__dict__.items():
            if isinstance(obj, _MagicTemplateMeta):
                new_attr = copy_class(obj, cls, name=name)
            elif (
                name.startswith("_") or isinstance(obj, _pass_type) or not callable(obj)
            ):
                # private method, non-action-like object, not-callable object are passed.
                new_attr = obj
            elif callable(obj) and get_additional_option(obj, "record", True):
                new_attr = _inject_recorder(obj)
            else:
                new_attr = obj

            _dict[name] = new_attr
    return _dict


def _define_popup(self: BaseGui, obj, widget: PushButtonPlus | Action):
    # deal with popup mode.
    popup_mode = self._popup_mode
    if popup_mode == PopUpMode.popup:
        # To be popped up correctly, window flags of FunctionGui should be
        # "windowFlags" and should appear at the center.
        def _prep(mgui: FunctionGui):
            mgui.native.setParent(self.native, mgui.native.windowFlags())
            move_to_screen_center(mgui.native)

    elif popup_mode == PopUpMode.parentlast:

        def _prep(mgui: FunctionGui):
            parent_self = self._search_parent_magicclass()
            parent_self.append(mgui)

    elif popup_mode == PopUpMode.first:

        def _prep(mgui: FunctionGui):
            child_self = _child_that_has_widget(self, obj, widget)
            child_self.insert(0, mgui)

    elif popup_mode == PopUpMode.last:

        def _prep(mgui: FunctionGui):
            child_self = _child_that_has_widget(self, obj, widget)
            child_self.append(mgui)

    elif popup_mode == PopUpMode.above:

        def _prep(mgui: FunctionGui):
            child_self = _child_that_has_widget(self, obj, widget)
            i = _get_index(child_self, widget)
            child_self.insert(i, mgui)

    elif popup_mode == PopUpMode.below:

        def _prep(mgui: FunctionGui):
            child_self = _child_that_has_widget(self, obj, widget)
            i = _get_index(child_self, widget)
            child_self.insert(i + 1, mgui)

    elif popup_mode == PopUpMode.dock:
        from .class_gui import MainWindowClassGui

        def _prep(mgui: FunctionGui):

            parent_self = self._search_parent_magicclass()
            viewer = parent_self.parent_viewer
            if viewer is None:
                if isinstance(parent_self, MainWindowClassGui):
                    parent_self.add_dock_widget(mgui)
                else:
                    msg = (
                        "Cannot add dock widget to a normal container. Please use\n"
                        ">>> @magicclass(widget_type='mainwindow')\n"
                        "to create main window widget, or add the container as a dock "
                        "widget in napari."
                    )
                    warnings.warn(msg, UserWarning)

            else:
                viewer.window.add_dock_widget(
                    mgui, name=_get_widget_name(widget), area="right"
                )

    elif popup_mode == PopUpMode.dialog:

        def _prep(mgui: FunctionGui):
            mgui.call_button.visible = False

    else:
        raise RuntimeError(popup_mode)
    return _prep


def _implement_confirmation(
    method: MethodType,
    self: BaseGui,
    text: str,
    condition: Callable[[BaseGui], bool] | str,
    callback: Callable[[str, BaseGui], None],
):
    """Implement confirmation callback to a method."""
    sig = inspect.signature(method)

    @wraps(method)
    def _method(*args, **kwargs):
        if self[method.__name__].running:
            arguments = sig.bind(*args, **kwargs)
            arguments.apply_defaults()
            all_args = arguments.arguments
            all_args.update(self=self)
            need_confirmation = False
            if isinstance(condition, str):
                try:
                    need_confirmation = eval(condition, {}, all_args)
                except Exception as e:
                    msg = e.args[0]
                    e.args = (
                        f"Exception happened on evaluating condition {condition!r}.\n"
                        f"{type(e).__name__}: {msg}"
                    )
                    raise e
            elif callable(condition):
                need_confirmation = condition(self)
            else:
                warnings.warn(
                    f"Condition {condition} should be callable or string but got type "
                    f"{type(condition)}. No confirmation was executed.",
                    UserWarning,
                )
            if need_confirmation:
                callback(text.format(**all_args), self)

        return method(*args, **kwargs)

    if hasattr(method, "__signature__"):
        _method.__signature__ = method.__signature__

    return _method


def _cleanup_tb(tb: TracebackType):
    """Remove useless info from a traceback object."""
    current_tb = tb
    while current_tb is not None:
        if current_tb.tb_frame.f_code.co_name == "_recordable":
            tb = current_tb.tb_next
            break
        current_tb = current_tb.tb_next
    return tb
